# coding=utf-8
"""Tools for getting and displaying the time."""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime

import pytz


# various time units measured in seconds; approximated for months and years
SECONDS = 1
MINUTES = 60 * SECONDS
HOURS = 60 * MINUTES
DAYS = 24 * HOURS
MONTHS = int(30.5 * DAYS)
YEARS = 365 * DAYS


def validate_timezone(zone):
    """Return an IETF timezone from the given IETF zone or common abbreviation.

    :param str zone: in a strict or a human-friendly format
    :return: the valid IETF timezone properly formatted
    :raise ValueError: when ``zone`` is not a valid timezone

    Prior to checking timezones, two transformations are made to make the zone
    names more human-friendly:

    1. the string is split on ``', '``, the pieces reversed, and then joined
       with ``/`` (*"New York, America"* becomes *"America/New York"*)
    2. Remaining spaces are replaced with ``_``

    This means ``new york, america`` becomes ``America/New_York``, and ``utc``
    becomes ``UTC``. In the majority of user-facing interactions, such
    case-insensitivity will be expected.

    If the zone is not valid, ``ValueError`` will be raised.
    """
    if zone is None:
        return None

    zone = '/'.join(reversed(zone.split(', '))).replace(' ', '_')
    try:
        tz = pytz.timezone(zone)
    except pytz.exceptions.UnknownTimeZoneError:
        raise ValueError('Invalid time zone.')
    return tz.zone


def validate_format(tformat):
    """Validate a time format string.

    :param str tformat: the format string to validate
    :return: the format string, if valid
    :raise ValueError: when ``tformat`` is not a valid time format string
    """
    try:
        time = datetime.datetime.utcnow()
        time.strftime(tformat)
    except (ValueError, TypeError):
        raise ValueError('Invalid time format.')
    return tformat


def get_nick_timezone(db, nick):
    """Get a nick's timezone from database.

    :param db: Bot's database handler (usually ``bot.db``)
    :type db: :class:`~sopel.db.SopelDB`
    :param nick: IRC nickname
    :type nick: :class:`~sopel.tools.Identifier`
    :return: the timezone associated with the ``nick``

    If a timezone cannot be found for ``nick``, or if it is invalid, ``None``
    will be returned.
    """
    try:
        return validate_timezone(db.get_nick_value(nick, 'timezone'))
    except ValueError:
        return None


def get_channel_timezone(db, channel):
    """Get a channel's timezone from database.

    :param db: Bot's database handler (usually ``bot.db``)
    :type db: :class:`~sopel.db.SopelDB`
    :param channel: IRC channel name
    :type channel: :class:`~sopel.tools.Identifier`
    :return: the timezone associated with the ``channel``

    If a timezone cannot be found for ``channel``, or if it is invalid,
    ``None`` will be returned.
    """
    try:
        return validate_timezone(db.get_channel_value(channel, 'timezone'))
    except ValueError:
        return None


def get_timezone(db=None, config=None, zone=None, nick=None, channel=None):
    """Find, and return, the appropriate timezone.

    :param db: bot database object (optional)
    :type db: :class:`~.db.SopelDB`
    :param config: bot config object (optional)
    :type config: :class:`~.config.Config`
    :param str zone: preferred timezone name (optional)
    :param str nick: nick whose timezone to use, if set (optional)
    :param str channel: channel whose timezone to use, if set (optional)

    Timezone is pulled in the following priority:

    1. ``zone``, if it is valid
    2. The timezone for the channel or nick ``zone`` in ``db`` if one is set
       and valid.
    3. The timezone for the nick ``nick`` in ``db``, if one is set and valid.
    4. The timezone for the channel ``channel`` in ``db``, if one is set and
       valid.
    5. The default timezone in ``config``, if one is set and valid.

    If ``db`` is not given, or given but not set up, steps 2 and 3 will be
    skipped. If ``config`` is not given, step 4 will be skipped. If no step
    yields a valid timezone, ``None`` is returned.

    Valid timezones are those present in the IANA Time Zone Database.

    .. seealso::

       The :func:`validate_timezone` function handles the validation and
       formatting of the timezone.

    """
    def _check(zone):
        try:
            return validate_timezone(zone)
        except ValueError:
            return None

    tz = None

    if zone:
        tz = _check(zone)
        if not tz:
            tz = _check(
                db.get_nick_or_channel_value(zone, 'timezone'))
    if not tz and nick:
        tz = _check(db.get_nick_value(nick, 'timezone'))
    if not tz and channel:
        tz = _check(db.get_channel_value(channel, 'timezone'))
    if not tz and config and config.core.default_timezone:
        tz = _check(config.core.default_timezone)
    return tz


def format_time(db=None, config=None, zone=None, nick=None, channel=None,
                time=None):
    """Return a formatted string of the given time in the given zone.

    :param db: bot database object (optional)
    :type db: :class:`~.db.SopelDB`
    :param config: bot config object (optional)
    :type config: :class:`~.config.Config`
    :param str zone: name of timezone to use (optional)
    :param str nick: nick whose time format to use, if set (optional)
    :param str channel: channel whose time format to use, if set (optional)
    :param time: the time value to format (optional)
    :type time: :class:`~datetime.datetime`

    ``time``, if given, should be a naive ``datetime.datetime`` object and will
    be treated as being in the UTC timezone. If it is not given, the current
    time will be used. If ``zone`` is given it must be present in the IANA Time
    Zone Database; ``get_timezone`` can be helpful for this. If ``zone`` is not
    given, UTC will be assumed.

    The format for the string is chosen in the following order:

    1. The format for the nick ``nick`` in ``db``, if one is set and valid.
    2. The format for the channel ``channel`` in ``db``, if one is set and
       valid.
    3. The default format in ``config``, if one is set and valid.
    4. ISO-8601

    If ``db`` is not given or is not set up, steps 1 and 2 are skipped. If
    ``config`` is not given, step 3 will be skipped.
    """
    tformat = None
    if db:
        if nick:
            tformat = db.get_nick_value(nick, 'time_format')
        if not tformat and channel:
            tformat = db.get_channel_value(channel, 'time_format')
    if not tformat and config and config.core.default_time_format:
        tformat = config.core.default_time_format
    if not tformat:
        tformat = '%Y-%m-%d - %T%Z'

    if not time:
        time = datetime.datetime.utcnow()

    if not zone:
        return time.strftime(tformat)
    else:
        if not time.tzinfo:
            utc = pytz.timezone('UTC')
            time = utc.localize(time)
        zone = pytz.timezone(zone)
        return time.astimezone(zone).strftime(tformat)


def seconds_to_split(seconds):
    """Split an amount of ``seconds`` into years, months, days, etc.

    :param int seconds: amount of time in seconds
    :return: the time split into a tuple of years, months, days, hours,
             minutes, and seconds
    :rtype: :class:`tuple`

    Examples::

        >>> seconds_to_split(7800)
        (0, 0, 0, 2, 10, 0)
        >>> seconds_to_split(143659)
        (0, 0, 1, 15, 54, 19)

    """
    years, seconds_left = divmod(int(seconds), YEARS)
    months, seconds_left = divmod(seconds_left, MONTHS)
    days, seconds_left = divmod(seconds_left, DAYS)
    hours, seconds_left = divmod(seconds_left, HOURS)
    minutes, seconds_left = divmod(seconds_left, MINUTES)

    return years, months, days, hours, minutes, seconds_left


def get_time_unit(years=0, months=0, days=0, hours=0, minutes=0, seconds=0):
    """Map a time in (y, m, d, h, min, s) to its labels.

    :param int years: number of years
    :param int months: number of months
    :param int days: number of days
    :param int hours: number of hours
    :param int minutes: number of minutes
    :param int seconds: number of seconds
    :return: a tuple of 2-value tuples, each for a time amount and its label
    :rtype: :class:`tuple`

    This helper function get a time split in years, months, days, hours,
    minutes, and seconds to return a tuple with the correct label for each
    unit. The label is pluralized and account for zéro, one, and more than one
    value per unit::

        >>> get_time_unit(days=1, hours=15, minutes=54, seconds=19)
        (
            (0, 'years'),
            (0, 'months'),
            (1, 'day'),
            (15, 'hours'),
            (54, 'minutes'),
            (19, 'seconds'),
        )

    This function can be used with :func:`seconds_to_split`::

        >>> get_time_unit(*seconds_to_split(143659))
        # ... same result as the example above

    .. note::

        This function always returns a tuple with **all** time units, even when
        their amount is 0 (which is their default value).

    """
    years_text = "year{}".format("s" if years != 1 else "")
    months_text = "month{}".format("s" if months != 1 else "")
    days_text = "day{}".format("s" if days != 1 else "")
    hours_text = "hour{}".format("s" if hours != 1 else "")
    minutes_text = "minute{}".format("s" if minutes != 1 else "")
    seconds_text = "second{}".format("s" if seconds != 1 else "")

    return (
        (years, years_text),
        (months, months_text),
        (days, days_text),
        (hours, hours_text),
        (minutes, minutes_text),
        (seconds, seconds_text),
    )


def seconds_to_human(secs, granularity=2):
    """Format :class:`~datetime.timedelta` as a human-readable relative time.

    :param secs: time difference to format
    :type secs: :class:`~datetime.timedelta` or integer
    :param int granularity: number of time units to return (default to 2)

    Inspiration for function structure from:
    https://gist.github.com/Highstaker/280a09591df4a5fb1363b0bbaf858f0d

    Examples::

        >>> seconds_to_human(65707200)
        '2 years, 1 month ago'
        >>> seconds_to_human(-17100)  # negative amount
        'in 4 hours, 45 minutes'
        >>> seconds_to_human(-709200)
        'in 8 days, 5 hours'
        >>> seconds_to_human(39441600, 1)  # 1 year + 3 months
        '1 year ago'

    This function can be used with a :class:`~datetime.timedelta`::

        >>> from datetime import timedelta
        >>> seconds_to_human(timedelta(days=42, seconds=278))
        '1 month, 11 days ago'

    The ``granularity`` argument controls how detailed the result is::

        >>> seconds_to_human(3672)  # 2 by default
        '1 hour, 1 minute ago'
        >>> seconds_to_human(3672, granularity=3)
        '1 hour, 1 minute, 12 seconds ago'
        >>> seconds_to_human(3672, granularity=1)
        '1 hour ago'

    """
    if isinstance(secs, datetime.timedelta):
        secs = secs.total_seconds()

    future = False
    if secs < 0:
        future = True

    secs = int(secs)
    secs = abs(secs)

    if secs == 0:
        # zero is a special case that the algorithm below won't handle correctly (#1841)
        result = "0 seconds"
    else:
        result = ", ".join([
            "%s %s" % (value, unit)
            for value, unit in get_time_unit(*seconds_to_split(secs))
            if value
        ][:granularity])

    if future is False:
        result += " ago"
    else:
        result = "in " + result
    return result



# coding=utf-8
"""
search.py - Sopel Search Engine Plugin
Copyright 2008-9, Sean B. Palmer, inamidst.com
Copyright 2012, Elsie Powell, embolalia.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import re

import requests
import xmltodict

from sopel import plugin
from sopel.tools import web

PLUGIN_OUTPUT_PREFIX = '[search] '

header_spoof = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36'
}
r_bing = re.compile(r'<li(?: class="b_algo")?><h2><a href="([^"]+)"')
r_duck = re.compile(r'nofollow" class="[^"]+" href="(?!(?:https?:\/\/r\.search\.yahoo)|(?:https?:\/\/duckduckgo\.com\/y\.js)(?:\/l\/\?kh=-1&amp;uddg=))(.*?)">')


def bing_search(query, lang='en-US'):
    base = 'https://www.bing.com/search'
    parameters = {
        'mkt': lang,
        'q': query,
    }
    response = requests.get(base, parameters, headers=header_spoof)
    m = r_bing.search(response.text)
    if m:
        return m.group(1)


def duck_search(query):
    query = query.replace('!', '')
    base = 'https://duckduckgo.com/html/'
    parameters = {
        'kl': 'us-en',
        'q': query,
    }
    bytes = requests.get(base, parameters, headers=header_spoof).text
    if 'web-result' in bytes:  # filter out the adds on top of the page
        bytes = bytes.split('web-result')[1]
    m = r_duck.search(bytes)
    if m:
        return web.decode(m.group(1))


def duck_api(query):
    if '!bang' in query.lower():
        return 'https://duckduckgo.com/bang.html'

    base = 'https://api.duckduckgo.com/'
    parameters = {
        'format': 'json',
        'no_html': '1',
        'no_redirect': '1',
        'q': query,
    }
    try:
        results = requests.get(base, parameters).json()
    except ValueError:
        return None
    if results['Redirect']:
        return results['Redirect']
    else:
        return None


@plugin.command('duck', 'ddg', 'g')
# test for bad Unicode handling in py2
@plugin.example(
    '.duck site:grandorder.wiki chulainn alter',
    r'https?:\/\/grandorder\.wiki\/C%C3%BA_Chulainn.*',
    re=True,
    online=True,
    vcr=True)
# the last example (in source line order) is what .help displays
@plugin.example(
    '.duck sopel.chat irc bot',
    r'https?:\/\/.*sopel.*',
    re=True,
    online=True,
    vcr=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def duck(bot, trigger):
    """Queries DuckDuckGo for the specified input."""
    query = trigger.group(2)
    if not query:
        bot.reply('{prefix}{command} what?'.format(
            prefix=bot.settings.core.help_prefix,
            command=trigger.group(1),
        ))
        return

    # If the API gives us something, say it and stop
    result = duck_api(query)
    if result:
        bot.say(result)
        return

    # Otherwise, look it up on the HTML version
    uri = duck_search(query)

    if uri:
        bot.say(uri)
        if 'last_seen_url' in bot.memory:
            bot.memory['last_seen_url'][trigger.sender] = uri
    else:
        msg = "No results found for '%s'." % query
        if query.count('site:') >= 2:
            # This check exists because of issue #1415. The git.io link will take the user there.
            # (Better a sopel.chat link, but it's not set up to do that. This is shorter anyway.)
            msg += " Try again with at most one 'site:' operator. See https://git.io/fpKtP for why."
        bot.reply(msg)


@plugin.command('bing')
@plugin.example('.bing sopel.chat irc bot')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def bing(bot, trigger):
    """Queries Bing for the specified input."""
    if not trigger.group(2):
        bot.reply('{}bing what?'.format(bot.settings.core.help_prefix))
        return
    query = trigger.group(2)
    result = bing_search(query)
    if result:
        bot.say(result)
    else:
        bot.reply("No results found for '%s'." % query)


@plugin.command('search')
@plugin.example('.search sopel irc bot')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def search(bot, trigger):
    """Searches both Bing and DuckDuckGo."""
    if not trigger.group(2):
        bot.reply('{}search for what?'.format(bot.settings.core.help_prefix))
        return
    query = trigger.group(2)
    bu = bing_search(query) or '-'
    du = duck_search(query) or '-'

    if bu == du:
        result = '%s (b, d)' % bu
    else:
        if len(bu) > 150:
            bu = '(extremely long link)'
        if len(du) > 150:
            du = '(extremely long link)'
        result = '%s (b), %s (d)' % (bu, du)

    bot.say(result)


@plugin.command('suggest')
@plugin.example('.suggest wikip', 'wikipedia', online=True, vcr=True)
@plugin.example('.suggest', '.suggest what?')
@plugin.example('.suggest lkashdfiauwgaef', 'Sorry, no result.', online=True, vcr=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def suggest(bot, trigger):
    """Suggests terms starting with given input"""
    if not trigger.group(2):
        bot.reply('{}suggest what?'.format(bot.settings.core.help_prefix))
        return
    query = trigger.group(2)
    # Using Google isn't necessarily ideal, but at most they'll be able to build
    # a composite profile of all users on a given instance, not a profile of any
    # single user. This can be switched out as soon as someone finds (or builds)
    # an alternative suggestion API.
    base = 'https://suggestqueries.google.com/complete/search'
    parameters = {
        'output': 'toolbar',
        'hl': 'en',
        'q': query,
    }
    response = requests.get(base, parameters)
    answer = xmltodict.parse(response.text)['toplevel']
    try:
        answer = answer['CompleteSuggestion'][0]['suggestion']['@data']
    except TypeError:
        answer = None
    if answer:
        bot.say(answer)
    else:
        bot.reply('Sorry, no result.')



# coding=utf-8
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import sys

from dns import rdtypes, resolver

from sopel.tools import deprecated

if sys.version_info.major >= 3:
    unicode = str


MYINFO_ARGS = ['client', 'servername', 'version']


def get_cnames(domain):
    """Determine the CNAMEs for a given domain.

    :param str domain: domain to check
    :return: list (of str)
    """
    try:
        answer = resolver.query(domain, "CNAME")
    except resolver.NoAnswer:
        return []

    return [
        data.to_text()[:-1]
        for data in answer
        if isinstance(data, rdtypes.ANY.CNAME.CNAME)
    ]


def safe(string):
    """Remove newlines from a string.

    :param str string: input text to process
    :return: the string without newlines
    :rtype: str
    :raises TypeError: when ``string`` is ``None``

    This function removes newlines from a string and always returns a unicode
    string (as in ``str`` on Python 3 and ``unicode`` on Python 2), but doesn't
    strip or alter it in any other way::

        >>> safe('some text\\r\\n')
        'some text'

    This is useful to ensure a string can be used in a IRC message.

    .. versionchanged:: 7.1

        This function now raises a :exc:`TypeError` instead of an unpredictable
        behaviour when given ``None``.

    """
    if string is None:
        raise TypeError('safe function requires a string, not NoneType')
    if sys.version_info.major >= 3 and isinstance(string, bytes):
        string = string.decode("utf8")
    elif sys.version_info.major < 3:
        if not isinstance(string, unicode):
            string = unicode(string, encoding='utf8')
    string = string.replace('\n', '')
    string = string.replace('\r', '')
    return string


class CapReq(object):
    """Represents a pending CAP REQ request.

    :param str prefix: either ``=`` (must be enabled),
                       ``-`` (must **not** be enabled),
                       or empty string (desired but optional)
    :param str plugin: the requesting plugin's name
    :param failure: function to call if this capability request fails
    :type failure: :term:`function`
    :param str arg: optional capability value; the request will fail if
                    the server's value is different
    :param success: function to call if this capability request succeeds
    :type success: :term:`function`

    The ``success`` and ``failure`` callbacks must accept two arguments:
    ``bot`` (a :class:`~sopel.bot.Sopel` instance) and ``cap`` (the name of
    the requested capability, as a string).

    .. seealso::
        For more information on how capability requests work, see the
        documentation for :meth:`sopel.irc.AbstractBot.cap_req`.
    """
    def __init__(self, prefix, plugin, failure=None, arg=None, success=None):
        def nop(bot, cap):
            pass
        # TODO at some point, reorder those args to be sane
        self.prefix = prefix
        self.plugin = plugin
        self.arg = arg
        self.failure = failure or nop
        self.success = success or nop

    @property
    @deprecated(
        reason='use the `plugin` property instead',
        version='7.1',
        removed_in='8.0',
    )
    def module(self):
        return self.plugin


class MyInfo(collections.namedtuple('MyInfo', MYINFO_ARGS)):
    """Store client, servername, and version from ``RPL_MYINFO`` events.

    .. seealso::

        https://modern.ircdocs.horse/#rplmyinfo-004

    """
    # TODO: replace by a class using typing.NamedTuple (new in Python 3.5+)
    # probably in Sopel 8.0 (due to drop most old Python versions)



# coding=utf-8
"""
xkcd.py - Sopel xkcd Plugin
Copyright 2010, Michael Yanovich (yanovich.net), and Morgan Goose
Copyright 2012, Lior Ramati
Copyright 2013, Elsie Powell (embolalia.com)
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import random
import re

import requests

from sopel import plugin
from sopel.modules.search import bing_search

PLUGIN_OUTPUT_PREFIX = '[xkcd] '

ignored_sites = [
    # For searching the web
    'almamater.xkcd.com',
    'blog.xkcd.com',
    'blag.xkcd.com',
    'forums.xkcd.com',
    'fora.xkcd.com',
    'forums3.xkcd.com',
    'store.xkcd.com',
    'wiki.xkcd.com',
    'what-if.xkcd.com',
]
sites_query = ' site:xkcd.com -site:' + ' -site:'.join(ignored_sites)


def get_info(number=None):
    if number:
        url = 'https://xkcd.com/{}/info.0.json'.format(number)
    else:
        url = 'https://xkcd.com/info.0.json'
    data = requests.get(url).json()
    data['url'] = 'https://xkcd.com/' + str(data['num'])
    return data


def web_search(query):
    url = bing_search(query + sites_query)
    if not url:
        return None
    match = re.match(r'(?:https?://)?xkcd.com/(\d+)/?', url)
    if match:
        return match.group(1)


@plugin.command('xkcd')
@plugin.example(".xkcd 1782", user_help=True)
@plugin.example(".xkcd", user_help=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def xkcd(bot, trigger):
    """Finds an xkcd comic strip.

    Takes one of 3 inputs:

      * If no input is provided it will return a random comic
      * If numeric input is provided it will return that comic, or the
        nth-latest comic if the number is non-positive
      * If non-numeric input is provided it will return the first search result
        for those keywords on the xkcd.com site
    """
    # get latest comic for rand function and numeric input
    latest = get_info()
    max_int = latest['num']

    # if no input is given (pre - lior's edits code)
    if not trigger.group(2):  # get rand comic
        random.seed()
        requested = get_info(random.randint(1, max_int + 1))
    else:
        query = trigger.group(2).strip()

        numbered = re.match(r"^(#|\+|-)?(\d+)$", query)
        if numbered:
            query = int(numbered.group(2))
            if numbered.group(1) == "-":
                query = -query
            return numbered_result(bot, query, latest)
        else:
            # Non-number: search the web.
            if (query.lower() == "latest" or query.lower() == "newest"):
                requested = latest
            else:
                number = web_search(query)
                if not number:
                    bot.reply('Could not find any comics for that query.')
                    return
                requested = get_info(number)

    say_result(bot, requested)


def numbered_result(bot, query, latest, commanded=True):
    max_int = latest['num']
    if query > max_int:
        bot.reply(("Sorry, comic #{} hasn't been posted yet. "
                   "The last comic was #{}").format(query, max_int))
        return
    elif query <= -max_int:
        bot.reply(("Sorry, but there were only {} comics "
                   "released yet so far").format(max_int))
        return
    elif abs(query) == 0:
        requested = latest
    elif query == 404 or max_int + query == 404:
        bot.say("404 - Not Found")  # don't error on that one
        return
    elif query > 0:
        requested = get_info(query)
    else:
        # Negative: go back that many from current
        requested = get_info(max_int + query)

    say_result(bot, requested, commanded)


def say_result(bot, result, commanded=True):
    parts = [
        result['title'],
        'Alt-text: ' + result['alt'],
    ]

    if commanded:
        parts.append(result['url'])

    bot.say(' | '.join(parts))


@plugin.url(r'xkcd.com/(\d+)')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def get_url(bot, trigger, match):
    latest = get_info()
    numbered_result(bot, int(match.group(1)), latest, commanded=False)


@plugin.url(r'https?://xkcd\.com/?$')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def xkcd_main_page(bot, trigger, match):
    latest = get_info()
    numbered_result(bot, 0, latest, commanded=False)



# coding=utf-8
"""
announce.py - Sopel Announcement Plugin
Sends announcements to all channels the bot has joined.
Copyright © 2013, Elad Alfassa, <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import itertools

from sopel import plugin


def _chunks(items, size):
    """Break a list of items into groups.

    :param items: the collection of items to chunk
    :type items: :term:`iterable`
    :param int size: the size of each chunk
    :return: a :term:`generator` of chunks
    :rtype: :term:`generator` of :class:`tuple`
    """
    # This approach is safer than slicing with non-subscriptable types,
    # for example `dict_keys` objects
    iterator = iter(items)
    # TODO: Simplify to assignment expression (`while cond := expr`)
    # when dropping Python 3.7
    chunk = tuple(itertools.islice(iterator, size))
    while chunk:
        yield chunk
        chunk = tuple(itertools.islice(iterator, size))


@plugin.command('announce')
@plugin.example('.announce Some important message here')
@plugin.require_admin('Sorry, I can\'t let you do that', reply=True)
@plugin.output_prefix('[ANNOUNCEMENT] ')
def announce(bot, trigger):
    """Send an announcement to all channels the bot is in."""
    if trigger.group(2) is None:
        bot.reply('Announce what? I need a message to say.')
        return

    size = 1
    try:
        size = bot.isupport.TARGMAX.get('PRIVMSG', size)
    except AttributeError:
        pass

    channels = _chunks(bot.channels.keys(), size)
    for cgroup in channels:
        bot.say(trigger.group(2), ','.join(cgroup))

    bot.reply('Announce complete.')



# coding=utf-8
"""Test mocks: they fake objects for testing.

.. versionadded:: 7.0
"""
from __future__ import absolute_import, division, print_function, unicode_literals


from sopel.irc.abstract_backends import AbstractIRCBackend


class MockIRCBackend(AbstractIRCBackend):
    """Fake IRC connection backend for testing purpose.

    :param bot: a Sopel instance
    :type bot: :class:`sopel.bot.Sopel`

    This backend doesn't require an actual connection. Instead, it stores every
    message sent in the :attr:`message_sent` list.

    You can use the :func:`~sopel.tests.rawlist` function to compare the
    messages easily, and the :meth:`clear_message_sent` method to clear
    previous messages::

        >>> from sopel.tests import rawlist, mocks
        >>> backend = mocks.MockIRCBackend(bot=None)
        >>> backend.irc_send(b'PRIVMSG #channel :Hi!\\r\\n')
        >>> backend.message_sent == rawlist('PRIVMSG #channel :Hi!')
        True
        >>> backend.clear_message_sent()
        [b'PRIVMSG #channel :Hi!\\r\\n']
        >>> backend.message_sent
        []

    .. seealso::

        The
        :class:`parent class <sopel.irc.abstract_backends.AbstractIRCBackend>`
        contains all the methods that can be used on this test backend.

    """
    def __init__(self, *args, **kwargs):
        super(MockIRCBackend, self).__init__(*args, **kwargs)
        self.message_sent = []
        """List of raw messages sent by the bot.

        This list will be populated each time the :meth:`irc_send` method is
        used: it will contain the raw IRC lines the bot wanted to send.

        You can clear this list with the :meth:`clear_message_sent` method, or
        use the :func:`~sopel.tests.rawlist` function to compare it.
        """
        self.connected = False
        """Convenient status flag.

        Set to ``True`` to make the bot think it is connected.
        """

    def is_connected(self):
        return self.connected

    def irc_send(self, data):
        """Store ``data`` into :attr:`message_sent`."""
        self.message_sent.append(data)

    def clear_message_sent(self):
        """Clear and return previous messages sent.

        :return: a copy of the cleared messages sent
        :rtype: :class:`list`

        .. versionadded:: 7.1
        """
        # make a copy
        sent = list(self.message_sent)
        # clear the message sent
        self.message_sent = []
        return sent


class MockIRCServer(object):
    """Fake IRC Server that can send messages to a test bot.

    :param bot: test bot instance to send messages to
    :type bot: :class:`sopel.bot.Sopel`
    :param bool join_threads: whether message functions should join running
                              threads before returning (default: ``True``)

    This mock object helps developers when they want to simulate an IRC server
    sending messages to the bot.

    The default ``join_threads`` behavior is suitable for testing most common
    plugin callables, and ensures that all callables dispatched by the ``bot``
    in response to messages sent via this ``MockIRCServer`` are finished
    running before execution can continue. If set to ``False``, the mock
    server will not wait for the bot to finish processing threaded
    :term:`callables <Plugin callable>` before returning.

    .. note::

        You can override ``join_threads`` on a per-method-call basis with the
        ``blocking`` arguments to the instance methods below.

    The :class:`~sopel.tests.factories.IRCFactory` factory can be used to
    create such mock object, either directly or by using ``py.test`` and the
    :func:`~sopel.tests.pytest_plugin.ircfactory` fixture.

    .. versionadded:: 7.1

        The ``join_threads`` parameter.
    """
    def __init__(self, bot, join_threads=True):
        self.bot = bot
        self.join_threads = join_threads
        # TODO: `blocking` method args below should be made kwarg-ONLY in py3

    @property
    def chanserv(self):
        """ChanServ's message prefix."""
        return 'ChanServ!ChanServ@services.'

    def channel_joined(self, channel, users=None, blocking=None):
        """Send events as if the bot just joined a channel.

        :param str channel: channel to send message for
        :param list users: list (or tuple) of nicknames that will be present
                           in the ``RPL_NAMREPLY`` event
        :param bool blocking: whether to block until all triggered threads
                              have finished (optional)

        This will send 2 messages to the bot:

        * a ``RPL_NAMREPLY`` event (353), giving information about ``users``
          present in ``channel``
        * a ``RPL_ENDOFNAMES`` event (366) for completion

        Use this to emulate when the bot joins a channel, and the server
        replies with the list of connected users::

            factory.channel_joined('#test', ['Owner', '@ChanServ'])

        In this example, the bot will know that there are 2 other users present
        in ``#test``: "Owner" (a regular user) and "ChanServ" (which is a
        channel operator). Note that the bot itself will be added to the list
        of users automatically, and you **should not** pass it in the ``users``
        parameter.

        This is particularly useful to populate the bot's memory of who is in
        a channel.

        If ``blocking`` is ``True``, this method will wait to join all running
        triggers' threads before returning. Setting it to ``False`` will skip
        this step. If not specified, this :class:`MockIRCServer` instance's
        ``join_threads`` argument will be obeyed.

        .. versionadded:: 7.1

            The ``blocking`` parameter.

        .. seealso::

            The ``join_threads`` argument to :class:`MockIRCServer`.

        .. note::

            To add a user to a channel after using this method, you should
            use the :meth:`join` method.
        """
        # automatically add the bot's nick to the list
        users = set(users or [])
        users.add(self.bot.nick)
        message = ':irc.example.com 353 {bot} = {channel} :{users}'.format(
            bot=self.bot.nick,
            users=' '.join(list(users)),
            channel=channel,
        )
        self.bot.on_message(message)

        message = (
            ':irc.example.com 366 {bot} = {channel} '
            ':End of /NAMES list.'
        ).format(
            bot=self.bot.nick,
            channel=channel,
        )
        self.bot.on_message(message)

        if (blocking is None and self.join_threads) or blocking:
            for t in self.bot.running_triggers:
                t.join()

    def mode_set(self, channel, flags, users, blocking=None):
        """Send a MODE event for a ``channel``

        :param str channel: channel receiving the MODE event
        :param str flags: MODE flags set
        :param list users: users getting the MODE flags
        :param bool blocking: whether to block until all triggered threads
                              have finished (optional)

        This will send a MODE message as if ``ChanServ`` added/removed channel
        modes for a set of ``users``. This method assumes the ``flags``
        parameter follows the `IRC specification for MODE`__::

            factory.mode_set('#test', '+vo-v', ['UserV', UserOP', 'UserAnon'])

        If ``blocking`` is ``True``, this method will wait to join all running
        triggers' threads before returning. Setting it to ``False`` will skip
        this step. If not specified, this :class:`MockIRCServer` instance's
        ``join_threads`` argument will be obeyed.

        .. versionadded:: 7.1

            The ``blocking`` parameter.

        .. seealso::

            The ``join_threads`` argument to :class:`MockIRCServer`.

        .. __: https://tools.ietf.org/html/rfc1459#section-4.2.3
        """
        message = ':{chanserv} MODE {channel} {flags} {users}'.format(
            chanserv=self.chanserv,
            channel=channel,
            flags=flags,
            users=' '.join(users),
        )
        self.bot.on_message(message)

        if (blocking is None and self.join_threads) or blocking:
            for t in self.bot.running_triggers:
                t.join()

    def join(self, user, channel, blocking=None):
        """Send a ``channel`` JOIN event from ``user``.

        :param user: factory for the user who joins the ``channel``
        :type user: :class:`MockUser`
        :param str channel: channel the ``user`` joined
        :param bool blocking: whether to block until all triggered threads
                              have finished (optional)

        This will send a ``JOIN`` message as if ``user`` just joined the
        channel::

            factory.join(MockUser('NewUser'), '#test')

        If ``blocking`` is ``True``, this method will wait to join all running
        triggers' threads before returning. Setting it to ``False`` will skip
        this step. If not specified, this :class:`MockIRCServer` instance's
        ``join_threads`` argument will be obeyed.

        .. versionadded:: 7.1

            The ``blocking`` parameter.

        .. seealso::

            The ``join_threads`` argument to :class:`MockIRCServer`.

        .. seealso::

            This function is a shortcut to call the bot with the result from
            the user factory's :meth:`~MockUser.join` method.
        """
        self.bot.on_message(user.join(channel))

        if (blocking is None and self.join_threads) or blocking:
            for t in self.bot.running_triggers:
                t.join()

    def say(self, user, channel, text, blocking=None):
        """Send a ``PRIVMSG`` to ``channel`` by ``user``.

        :param user: factory for the user who sends a message to ``channel``
        :type user: :class:`MockUser`
        :param str channel: recipient of the ``user``'s ``PRIVMSG``
        :param str text: content of the message sent to the ``channel``
        :param bool blocking: whether to block until all triggered threads
                              have finished (optional)

        This will send a ``PRIVMSG`` message as if ``user`` sent it to the
        ``channel``, and the server forwarded it to its clients::

            factory.say(MockUser('NewUser'), '#test', '.shrug')

        If ``blocking`` is ``True``, this method will wait to join all running
        triggers' threads before returning. Setting it to ``False`` will skip
        this step. If not specified, this :class:`MockIRCServer` instance's
        ``join_threads`` argument will be obeyed.

        .. versionadded:: 7.1

            The ``blocking`` parameter.

        .. seealso::

            The ``join_threads`` argument to :class:`MockIRCServer`.

        .. seealso::

            This function is a shortcut to call the bot with the result from
            the user's :meth:`~MockUser.privmsg` method.
        """
        self.bot.on_message(user.privmsg(channel, text))

        if (blocking is None and self.join_threads) or blocking:
            for t in self.bot.running_triggers:
                t.join()

    def pm(self, user, text, blocking=None):
        """Send a ``PRIVMSG`` to the bot by a ``user``.

        :param user: factory for the user object who sends a message
        :type user: :class:`MockUser`
        :param str text: content of the message sent to the bot
        :param bool blocking: whether to block until all triggered threads
                              have finished (optional)

        This will send a ``PRIVMSG`` message as forwarded by the server for
        a ``user`` sending it to the bot::

            factory.pm(MockUser('NewUser'), 'A private word.')

        If ``blocking`` is ``True``, this method will wait to join all running
        triggers' threads before returning. Setting it to ``False`` will skip
        this step. If not specified, this :class:`MockIRCServer` instance's
        ``join_threads`` argument will be obeyed.

        .. versionadded:: 7.1

            The ``blocking`` parameter.

        .. seealso::

            The ``join_threads`` argument to :class:`MockIRCServer`.

        .. seealso::

            This function is a shortcut to call the bot with the result from
            the user factory's :meth:`~MockUser.privmsg` method, using the
            bot's nick as recipient.
        """
        self.bot.on_message(user.privmsg(self.bot.nick, text))

        if (blocking is None and self.join_threads) or blocking:
            for t in self.bot.running_triggers:
                t.join()


class MockUser(object):
    """Fake user that can generate messages to send to a bot.

    :param str nick: nickname
    :param str user: IRC username
    :param str host: user's host

    The :class:`~sopel.tests.factories.UserFactory` factory can be used to
    create such mock object, either directly or by using ``py.test`` and the
    :func:`~sopel.tests.pytest_plugin.userfactory` fixture.
    """
    def __init__(self, nick=None, user=None, host=None):
        self.nick = nick or 'Test'
        self.user = user or self.nick.lower()
        self.host = host or 'example.com'

    @property
    def prefix(self):
        """User's hostmask as seen by other users on the server.

        When the server forwards a User's command, it uses this prefix.
        """
        return '{nick}!{user}@{host}'.format(
            nick=self.nick, user=self.user, host=self.host)

    def privmsg(self, recipient, text):
        """Generate a ``PRIVMSG`` command forwarded by a server for the user.

        :param str recipient: a channel name or the bot's nick
        :param str text: content of the message
        :return: a ``PRIVMSG`` command forwarded by the server as if it
                 originated from the user's hostmask
        :rtype: str
        """
        message = ':{prefix} PRIVMSG {recipient} :{text}\r\n'.format(
            prefix=self.prefix,
            recipient=recipient,
            text=text,
        )

        assert len(message.encode('utf-8')) <= 512, (
            'PRIVMSG command must NOT exceed the 512 bytes limit '
            '(\\r\\n included). Trying to send this command:\n`%r`' % message
        )

        return message

    def join(self, channel):
        """Generate a ``JOIN`` command forwarded by the server for the user.

        :param str channel: channel the user joined
        :return: the ``JOIN`` command the server sends to its clients present
                 in the same ``channel`` when the user joins it.
        :rtype: str
        """
        return ':{prefix} JOIN {channel}\r\n'.format(
            prefix=self.prefix,
            channel=channel)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
import logging
import os
import sys

from sopel import config, plugins, tools

# Allow clean import *
__all__ = [
    'enumerate_configs',
    'find_config',
    'add_common_arguments',
    'load_settings',
    'redirect_outputs',
    'wizard',
    'plugins_wizard',
    # colors
    'green',
    'yellow',
    'red',
]

LOGGER = logging.getLogger(__name__)

RESET = '\033[0m'
RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[33m'


def _colored(text, color, reset=True):
    text = color + text
    if reset:
        return text + RESET
    return text


def green(text, reset=True):
    """Add ANSI escape sequences to make the text green in terminal.

    :param str text: text to colorized in green
    :param bool reset: if the text color must be reset after (default ``True``)
    :return: text with ANSI escape sequences for green color
    :rtype: str
    """
    return _colored(text, GREEN, reset)


def yellow(text, reset=True):
    """Add ANSI escape sequences to make the text yellow in terminal.

    :param str text: text to colorized in yellow
    :param bool reset: if the text color must be reset after (default ``True``)
    :return: text with ANSI escape sequences for yellow color
    :rtype: str
    """
    return _colored(text, YELLOW, reset)


def red(text, reset=True):
    """Add ANSI escape sequences to make the text red in terminal.

    :param str text: text to colorized in red
    :param bool reset: if the text color must be reset after (default ``True``)
    :return: text with ANSI escape sequences for red color
    :rtype: str
    """
    return _colored(text, RED, reset)


def wizard(filename):
    """Global Configuration Wizard.

    :param str filename: name of the new file to be created
    :return: the created configuration object

    This wizard function helps the creation of a Sopel configuration file,
    with its core section and its plugins' sections.
    """
    configdir, basename = os.path.split(filename)
    if not basename:
        raise config.ConfigurationError(
            'Sopel requires a filename for its configuration, not a directory')

    try:
        if not os.path.isdir(configdir):
            print('Creating config directory at {}'.format(configdir))
            os.makedirs(configdir)
            print('Config directory created')
    except Exception:
        tools.stderr('There was a problem creating {}'.format(configdir))
        raise

    name, ext = os.path.splitext(basename)
    if not ext:
        # Always add .cfg if filename does not have an extension
        filename = os.path.join(configdir, name + '.cfg')
    elif ext != '.cfg':
        # It is possible to use a non-cfg file for Sopel
        # but the wizard does not allow it at the moment
        raise config.ConfigurationError(
            'Sopel uses ".cfg" as configuration file extension, not "%s".' % ext)

    settings = config.Config(filename, validate=False)

    print("Please answer the following questions "
          "to create your configuration file (%s):\n" % filename)
    config.core_section.configure(settings)
    if settings.option(
        'Would you like to see if there are any plugins '
        'that need configuring'
    ):
        _plugins_wizard(settings)

    try:
        settings.save()
    except Exception:  # TODO: Be specific
        tools.stderr("Encountered an error while writing the config file. "
                     "This shouldn't happen. Check permissions.")
        raise

    print("Config file written successfully!")
    return settings


def plugins_wizard(filename):
    """Plugins Configuration Wizard.

    :param str filename: path to an existing Sopel configuration
    :return: the configuration object

    This wizard function helps to configure plugins for an existing Sopel
    config file.
    """
    if not os.path.isfile(filename):
        raise config.ConfigurationNotFound(filename)

    settings = config.Config(filename, validate=False)
    _plugins_wizard(settings)

    try:
        settings.save()
    except Exception:  # TODO: Be specific
        tools.stderr("Encountered an error while writing the config file. "
                     "This shouldn't happen. Check permissions.")
        raise

    return settings


def _plugins_wizard(settings):
    usable_plugins = plugins.get_usable_plugins(settings)
    for plugin, is_enabled in usable_plugins.values():
        if not is_enabled:
            # Do not configure non-enabled plugins
            continue

        name = plugin.name
        try:
            _plugin_wizard(settings, plugin)
        except Exception as e:
            LOGGER.exception('Error loading %s: %s', name, e)


def _plugin_wizard(settings, plugin):
    plugin.load()
    prompt = "Configure {}".format(plugin.get_label())
    if plugin.has_configure() and settings.option(prompt):
        plugin.configure(settings)


def enumerate_configs(config_dir, extension='.cfg'):
    """List configuration files from ``config_dir`` with ``extension``.

    :param str config_dir: path to the configuration directory
    :param str extension: configuration file's extension (default to ``.cfg``)
    :return: a list of configuration filenames found in ``config_dir`` with
             the correct ``extension``
    :rtype: list

    Example::

        >>> from sopel import cli, config
        >>> os.listdir(config.DEFAULT_HOMEDIR)
        ['config.cfg', 'extra.ini', 'plugin.cfg', 'README']
        >>> cli.enumerate_configs(config.DEFAULT_HOMEDIR)
        ['config.cfg', 'plugin.cfg']
        >>> cli.enumerate_configs(config.DEFAULT_HOMEDIR, '.ini')
        ['extra.ini']

    """
    if not os.path.isdir(config_dir):
        return

    for item in os.listdir(config_dir):
        if item.endswith(extension):
            yield item


def find_config(config_dir, name, extension='.cfg'):
    """Build the absolute path for the given configuration file ``name``.

    :param str config_dir: path to the configuration directory
    :param str name: configuration file ``name``
    :param str extension: configuration file's extension (default to ``.cfg``)
    :return: the path of the configuration file, either in the current
             directory or from the ``config_dir`` directory

    This function tries different locations:

    * the current directory
    * the ``config_dir`` directory with the ``extension`` suffix
    * the ``config_dir`` directory without a suffix

    Example::

        >>> from sopel.cli import utils
        >>> from sopel import config
        >>> os.listdir()
        ['local.cfg', 'extra.ini']
        >>> os.listdir(config.DEFAULT_HOMEDIR)
        ['config.cfg', 'extra.ini', 'plugin.cfg', 'README']
        >>> utils.find_config(config.DEFAULT_HOMEDIR, 'local.cfg')
        'local.cfg'
        >>> utils.find_config(config.DEFAULT_HOMEDIR, 'local')
        '/home/username/.sopel/local'
        >>> utils.find_config(config.DEFAULT_HOMEDIR, 'config')
        '/home/username/.sopel/config.cfg'
        >>> utils.find_config(config.DEFAULT_HOMEDIR, 'extra', '.ini')
        '/home/username/.sopel/extra.ini'

    """
    if os.path.isfile(name):
        return os.path.abspath(name)
    name_ext = name + extension
    for filename in enumerate_configs(config_dir, extension):
        if name_ext == filename:
            return os.path.join(config_dir, name_ext)

    return os.path.join(config_dir, name)


def add_common_arguments(parser):
    """Add common and configuration-related arguments to a ``parser``.

    :param parser: Argument parser (or subparser)
    :type parser: argparse.ArgumentParser

    This function adds the common arguments for Sopel's command line tools.
    It adds the following arguments:

    * ``-c``/``--config``: the name of the Sopel config, or its absolute path
    * ``--config-dir``: the directory to scan for config files

    This can be used on an argument parser, or an argument subparser, to handle
    these cases::

        [sopel-command] -c [filename]
        [sopel-command] [action] -c [filename]
        [sopel-command] --config-dir [directory] -c [name]

    Then, when the parser parses the command line arguments, it will expose
    ``config`` and ``configdir`` options that can be used to find and load
    Sopel's settings.

    The default value for ``config`` is either the value of the environment
    variable ``SOPEL_CONFIG``, or the string ``default``.

    .. seealso::

        The :func:`sopel.cli.utils.load_settings` function uses an ``options``
        object from a parser configured with such arguments.

    """
    parser.add_argument(
        '-c', '--config',
        default=os.environ.get('SOPEL_CONFIG') or 'default',
        metavar='filename',
        dest='config',
        help=inspect.cleandoc("""
            Use a specific configuration file.
            A config name can be given and the configuration file will be
            found in Sopel's homedir (defaults to ``~/.sopel/default.cfg``).
            An absolute pathname can be provided instead to use an
            arbitrary location.
            When the ``SOPEL_CONFIG`` environment variable is set and not
            empty, it is used as the default value.
        """))
    parser.add_argument(
        '--config-dir',
        default=os.environ.get('SOPEL_CONFIG_DIR') or config.DEFAULT_HOMEDIR,
        dest='configdir',
        help=inspect.cleandoc("""
            Look for configuration files in this directory.
            By default, Sopel will search in ``~/.sopel``.
            When the ``SOPEL_CONFIG_DIR`` environment variable is set and not
            empty, it is used as the default value.
        """))


def load_settings(options):
    """Load Sopel's settings using the command line's ``options``.

    :param options: parsed arguments
    :return: sopel configuration
    :rtype: :class:`sopel.config.Config`
    :raise sopel.config.ConfigurationNotFound: raised when configuration file
                                               is not found
    :raise sopel.config.ConfigurationError: raised when configuration is
                                            invalid

    This function loads Sopel's settings from ``options.config``. This option's
    value should be from one of these sources:

    * given by the command line argument,
    * ``SOPEL_CONFIG`` environment variable, if the argument is not used,
    * otherwise it should default to ``default``,

    then it returns it as a :class:`~sopel.config.Config` object.

    If the configuration file can not be found, a
    :exc:`sopel.config.ConfigurationNotFound` error will be raised.

    .. note::

        This function expects that ``options`` exposes two attributes:
        ``config`` and ``configdir``.

        The :func:`sopel.cli.utils.add_common_arguments` function should be
        used to add these options to the argument parser. This function is also
        responsible for using the environment variable or the default value.

    """
    filename = find_config(options.configdir, options.config)

    if not os.path.isfile(filename):
        raise config.ConfigurationNotFound(filename=filename)

    return config.Config(filename)


@tools.deprecated(reason='Obsoleted by modernized logging.',
                  version='7.0',
                  removed_in='8.0')
def redirect_outputs(settings, is_quiet=False):
    """Redirect ``sys``'s outputs using Sopel's settings.

    :param settings: Sopel's configuration
    :type settings: :class:`sopel.config.Config`
    :param bool is_quiet: Optional, set to True to make Sopel's outputs quiet

    Both ``sys.stderr`` and ``sys.stdout`` are redirected to a logfile.

    .. deprecated:: 7.0

        Sopel now uses the built-in logging system for its output, and this
        function is now deprecated.

    """
    logfile = os.path.os.path.join(
        settings.core.logdir, settings.basename + '.stdio.log')
    sys.stderr = tools.OutputRedirect(logfile, True, is_quiet)
    sys.stdout = tools.OutputRedirect(logfile, False, is_quiet)


def get_many_text(items, one, two, many):
    """Get the right text based on the number of ``items``."""
    message = ''
    if not items:
        return message

    items_count = len(items)

    if items_count == 1:
        message = one.format(item=items[0])
    elif items_count == 2:
        message = two.format(first=items[0], second=items[1])
    else:
        left = ', '.join(items[:-1])
        last = items[-1]
        message = many.format(left=left, last=last)

    return message



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import functools

from sopel import plugin
from sopel.tools import Identifier


@functools.total_ordering
class User(object):
    """A representation of a user Sopel is aware of.

    :param nick: the user's nickname
    :type nick: :class:`~.tools.Identifier`
    :param str user: the user's local username ("user" in `user@host.name`)
    :param str host: the user's hostname ("host.name" in `user@host.name`)
    """
    def __init__(self, nick, user, host):
        assert isinstance(nick, Identifier)
        self.nick = nick
        """The user's nickname."""
        self.user = user
        """The user's local username."""
        self.host = host
        """The user's hostname."""
        self.channels = {}
        """The channels the user is in.

        This maps channel name :class:`~sopel.tools.Identifier`\\s to
        :class:`Channel` objects.
        """
        self.account = None
        """The IRC services account of the user.

        This relies on IRCv3 account tracking being enabled.
        """
        self.away = None
        """Whether the user is marked as away."""

    hostmask = property(lambda self: '{}!{}@{}'.format(self.nick, self.user,
                                                       self.host))
    """The user's full hostmask."""

    def __eq__(self, other):
        if not isinstance(other, User):
            return NotImplemented
        return self.nick == other.nick

    def __lt__(self, other):
        if not isinstance(other, User):
            return NotImplemented
        return self.nick < other.nick


@functools.total_ordering
class Channel(object):
    """A representation of a channel Sopel is in.

    :param name: the channel name
    :type name: :class:`~.tools.Identifier`
    """
    def __init__(self, name):
        assert isinstance(name, Identifier)
        self.name = name
        """The name of the channel."""
        self.users = {}
        """The users in the channel.

        This maps nickname :class:`~sopel.tools.Identifier`\\s to :class:`User`
        objects.
        """
        self.privileges = {}
        """The permissions of the users in the channel.

        This maps nickname :class:`~sopel.tools.Identifier`\\s to bitwise
        integer values. This can be compared to appropriate constants from
        :mod:`sopel.plugin`.
        """
        self.topic = ''
        """The topic of the channel."""

        self.modes = {}
        """The channel's modes.

        For type A modes (nick/address list), the value is a set. For type B
        (parameter) or C (parameter when setting), the value is a string. For
        type D, the value is ``True``.

        .. note::

            Type A modes may only contain changes the bot has observed. Sopel
            does not automatically populate all modes and lists.
        """

        self.last_who = None
        """The last time a WHO was requested for the channel."""

    def clear_user(self, nick):
        """Remove ``nick`` from this channel.

        :param nick: the nickname of the user to remove
        :type nick: :class:`~.tools.Identifier`

        Called after a user leaves the channel via PART, KICK, QUIT, etc.
        """
        user = self.users.pop(nick, None)
        self.privileges.pop(nick, None)
        if user is not None:
            user.channels.pop(self.name, None)

    def add_user(self, user, privs=0):
        """Add ``user`` to this channel.

        :param user: the new user to add
        :type user: :class:`User`
        :param int privs: privilege bitmask (see constants in
                          :mod:`sopel.plugin`)

        Called when a new user JOINs the channel.
        """
        assert isinstance(user, User)
        self.users[user.nick] = user
        self.privileges[user.nick] = privs or 0
        user.channels[self.name] = self

    def has_privilege(self, nick, privilege):
        """Tell if a user has a ``privilege`` level or above in this channel.

        :param str nick: a user's nick in this channel
        :param int privilege: privilege level to check
        :rtype: bool

        This method checks the user's privilege level in this channel, i.e. if
        it has this level or higher privileges::

            >>> channel.add_user(some_user, plugin.OP)
            >>> channel.has_privilege(some_user.nick, plugin.VOICE)
            True

        The ``nick`` argument can be either a :class:`str` or a
        :class:`sopel.tools.Identifier`. If the user is not in this channel,
        it will be considered as not having any privilege.

        .. seealso::

            There are other methods to check the exact privilege level of a
            user, such as :meth:`is_oper`, :meth:`is_owner`, :meth:`is_admin`,
            :meth:`is_op`, :meth:`is_halfop`, and :meth:`is_voiced`.

        .. important::

            Not all IRC networks support all privilege levels. If you intend
            for your plugin to run on any network, it is safest to rely only
            on the presence of standard modes: ``+v`` (voice) and ``+o`` (op).

        """
        return self.privileges.get(Identifier(nick), 0) >= privilege

    def is_oper(self, nick):
        """Tell if a user has the OPER (operator) privilege level.

        :param str nick: a user's nick in this channel
        :rtype: bool

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the OPER privilege level::

            >>> channel.add_user(some_user, plugin.OPER)
            >>> channel.is_oper(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.OPER | plugin.VOICE)
            >>> channel.is_oper(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        return self.privileges.get(Identifier(nick), 0) & plugin.OPER

    def is_owner(self, nick):
        """Tell if a user has the OWNER privilege level.

        :param str nick: a user's nick in this channel
        :rtype: bool

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the OWNER privilege level::

            >>> channel.add_user(some_user, plugin.OWNER)
            >>> channel.is_owner(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.OWNER | plugin.VOICE)
            >>> channel.is_owner(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        return self.privileges.get(Identifier(nick), 0) & plugin.OWNER

    def is_admin(self, nick):
        """Tell if a user has the ADMIN privilege level.

        :param str nick: a user's nick in this channel
        :rtype: bool

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the ADMIN privilege level::

            >>> channel.add_user(some_user, plugin.ADMIN)
            >>> channel.is_admin(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.ADMIN | plugin.VOICE)
            >>> channel.is_admin(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        return self.privileges.get(Identifier(nick), 0) & plugin.ADMIN

    def is_op(self, nick):
        """Tell if a user has the OP privilege level.

        :param str nick: a user's nick in this channel
        :rtype: bool

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the OP privilege level::

            >>> channel.add_user(some_user, plugin.OP)
            >>> channel.is_op(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.OP | plugin.VOICE)
            >>> channel.is_op(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        """
        return self.privileges.get(Identifier(nick), 0) & plugin.OP

    def is_halfop(self, nick):
        """Tell if a user has the HALFOP privilege level.

        :param str nick: a user's nick in this channel
        :rtype: bool

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the HALFOP privilege level::

            >>> channel.add_user(some_user, plugin.HALFOP)
            >>> channel.is_halfop(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.HALFOP | plugin.VOICE)
            >>> channel.is_halfop(some_user.nick)
            True
            >>> channel.is_voiced(some_user.nick)
            True

        .. important::

            Not all IRC networks support this privilege mode. If you are
            writing a plugin for public distribution, ensure your code behaves
            sensibly if only ``+v`` (voice) and ``+o`` (op) modes exist.

        """
        return self.privileges.get(Identifier(nick), 0) & plugin.HALFOP

    def is_voiced(self, nick):
        """Tell if a user has the VOICE privilege level.

        :param str nick: a user's nick in this channel
        :rtype: bool

        Unlike :meth:`has_privilege`, this method checks if the user has been
        explicitly granted the VOICE privilege level::

            >>> channel.add_user(some_user, plugin.VOICE)
            >>> channel.is_voiced(some_user.nick)
            True
            >>> channel.add_user(some_user, plugin.OP)
            >>> channel.is_voiced(some_user.nick)
            False

        Note that you can always have more than one privilege level::

            >>> channel.add_user(some_user, plugin.VOICE | plugin.OP)
            >>> channel.is_voiced(some_user.nick)
            True
            >>> channel.is_op(some_user.nick)
            True

        """
        return self.privileges.get(Identifier(nick), 0) & plugin.VOICE

    def rename_user(self, old, new):
        """Rename a user.

        :param old: the user's old nickname
        :type old: :class:`~.tools.Identifier`
        :param new: the user's new nickname
        :type new: :class:`~.tools.Identifier`

        Called on NICK events.
        """
        if old in self.users:
            self.users[new] = self.users.pop(old)
            self.users[new].nick = new
        if old in self.privileges:
            self.privileges[new] = self.privileges.pop(old)

    def __eq__(self, other):
        if not isinstance(other, Channel):
            return NotImplemented
        return self.name == other.name

    def __lt__(self, other):
        if not isinstance(other, Channel):
            return NotImplemented
        return self.name < other.name



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys

import pytest

from sopel import config
from sopel.config import types


FAKE_CONFIG = """
[core]
owner=dgw
homedir={homedir}
"""


MULTILINE_CONFIG = FAKE_CONFIG + """
[spam]
eggs = one, two, three, four, and a half
bacons = grilled
    burn out, 
    , greasy, fat, and tasty
cheeses =   
    cheddar
      reblochon   
  camembert
channels =
    "#sopel"
    &peculiar
# regular comment
    # python 3 only comment
    "#private"
    "#startquote
    &endquote"
    "&quoted"

[somesection]
is_defined = no
"""  # noqa (trailing whitespaces are intended)

TEST_CHANNELS = [
    '#sopel',
    '&peculiar',
    '#private',
    '"#startquote',  # start quote without end quote: kept
    '&endquote"',
    '"&quoted"',  # quoted, but no #: quotes kept
]

if sys.version_info.major < 3:
    # Python 2.7's ConfigParser interprets as comment
    # a line that starts with # or ;.
    # Python 3, on the other hand, allows comments to be indented.
    # As a result, the same config file will result in a different
    # config object depending on the Python version used.
    # TODO: Deprecated with Python 2.7.
    TEST_CHANNELS = [
        '#sopel',
        '&peculiar',
        '# python 3 only comment',  # indented lines cannot be comments in Py2
        '#private',
        '"#startquote',
        '&endquote"',
        '"&quoted"',
    ]


class FakeConfigSection(types.StaticSection):
    valattr = types.ValidatedAttribute('valattr')
    listattr = types.ListAttribute('listattr')
    choiceattr = types.ChoiceAttribute('choiceattr', ['spam', 'egg', 'bacon'])
    booleanattr = types.BooleanAttribute('booleanattr')
    booleanattr_true = types.BooleanAttribute('booleanattr', default=True)
    af_fileattr = types.FilenameAttribute('af_fileattr', relative=False, directory=False)
    ad_fileattr = types.FilenameAttribute('ad_fileattr', relative=False, directory=True)
    rf_fileattr = types.FilenameAttribute('rf_fileattr', relative=True, directory=False)
    rd_fileattr = types.FilenameAttribute('rd_fileattr', relative=True, directory=True)


class SpamSection(types.StaticSection):
    eggs = types.ListAttribute('eggs')
    bacons = types.ListAttribute('bacons', strip=False)
    cheeses = types.ListAttribute('cheeses')
    channels = types.ListAttribute('channels')


@pytest.fixture
def tmphomedir(tmpdir):
    sopel_homedir = tmpdir.join('.sopel')
    sopel_homedir.mkdir()
    sopel_homedir.join('test.tmp').write('')
    sopel_homedir.join('test.d').mkdir()
    return sopel_homedir


@pytest.fixture
def fakeconfig(tmphomedir):
    conf_file = tmphomedir.join('conf.cfg')
    conf_file.write(FAKE_CONFIG.format(homedir=tmphomedir.strpath))

    test_settings = config.Config(conf_file.strpath)
    test_settings.define_section('fake', FakeConfigSection)
    return test_settings


@pytest.fixture
def multi_fakeconfig(tmphomedir):
    conf_file = tmphomedir.join('conf.cfg')
    conf_file.write(MULTILINE_CONFIG.format(homedir=tmphomedir.strpath))

    test_settings = config.Config(conf_file.strpath)
    test_settings.define_section('fake', FakeConfigSection)
    test_settings.define_section('spam', SpamSection)
    return test_settings


def test_validated_string_when_none(fakeconfig):
    fakeconfig.fake.valattr = None
    assert fakeconfig.fake.valattr is None


def test_listattribute_when_empty(fakeconfig):
    fakeconfig.fake.listattr = []
    assert fakeconfig.fake.listattr == []


def test_listattribute_with_one_value(fakeconfig):
    fakeconfig.fake.listattr = ['foo']
    assert fakeconfig.fake.listattr == ['foo']


def test_listattribute_with_multiple_values(fakeconfig):
    fakeconfig.fake.listattr = ['egg', 'sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['egg', 'sausage', 'bacon']


def test_listattribute_with_value_containing_comma(fakeconfig):
    fakeconfig.fake.listattr = ['spam, egg, sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam, egg, sausage', 'bacon']


def test_listattribute_with_value_containing_nonescape_backslash(fakeconfig):
    fakeconfig.fake.listattr = ['spam', r'egg\sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', r'egg\sausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', r'egg\tacos', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', r'egg\tacos', 'bacon']


def test_listattribute_with_value_containing_standard_escape_sequence(fakeconfig):
    fakeconfig.fake.listattr = ['spam', 'egg\tsausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', 'egg\tsausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', 'egg\nsausage', 'bacon']
    assert fakeconfig.fake.listattr == [
        'spam', 'egg', 'sausage', 'bacon'
    ], 'Line break are always converted to new item'

    fakeconfig.fake.listattr = ['spam', 'egg\\sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', 'egg\\sausage', 'bacon']


def test_listattribute_with_value_ending_in_special_chars(fakeconfig):
    fakeconfig.fake.listattr = ['spam', 'egg', 'sausage\\', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', 'egg', 'sausage\\', 'bacon']

    fakeconfig.fake.listattr = ['spam', 'egg', 'sausage,', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', 'egg', 'sausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', 'egg', 'sausage,,', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', 'egg', 'sausage', 'bacon']


def test_listattribute_with_value_containing_adjacent_special_chars(fakeconfig):
    fakeconfig.fake.listattr = ['spam', r'egg\,sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', r'egg\,sausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', r'egg\,\sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', r'egg\,\sausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', r'egg,\,sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', r'egg,\,sausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', 'egg,,sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', 'egg,,sausage', 'bacon']

    fakeconfig.fake.listattr = ['spam', r'egg\\sausage', 'bacon']
    assert fakeconfig.fake.listattr == ['spam', r'egg\\sausage', 'bacon']


def test_choiceattribute_when_none(fakeconfig):
    fakeconfig.fake.choiceattr = None
    assert fakeconfig.fake.choiceattr is None


def test_choiceattribute_when_not_in_set(fakeconfig):
    with pytest.raises(ValueError):
        fakeconfig.fake.choiceattr = 'sausage'


def test_choiceattribute_when_valid(fakeconfig):
    fakeconfig.fake.choiceattr = 'bacon'
    assert fakeconfig.fake.choiceattr == 'bacon'


def test_booleanattribute_default(fakeconfig):
    assert fakeconfig.fake.booleanattr is False
    assert fakeconfig.fake.booleanattr_true is True


def test_fileattribute_valid_absolute_file_path(fakeconfig):
    testfile = os.path.join(fakeconfig.core.homedir, 'test.tmp')
    fakeconfig.fake.af_fileattr = testfile
    assert fakeconfig.fake.af_fileattr == testfile


def test_fileattribute_valid_absolute_dir_path(fakeconfig):
    testdir = os.path.join(fakeconfig.core.homedir, 'test.d')
    fakeconfig.fake.ad_fileattr = testdir
    assert fakeconfig.fake.ad_fileattr == testdir


def test_fileattribute_given_relative_when_absolute(fakeconfig):
    with pytest.raises(ValueError):
        fakeconfig.fake.af_fileattr = '../testconfig.tmp'


def test_fileattribute_given_absolute_when_relative(fakeconfig):
    testfile = os.path.join(fakeconfig.core.homedir, 'test.tmp')
    fakeconfig.fake.rf_fileattr = testfile
    assert fakeconfig.fake.rf_fileattr == testfile


def test_fileattribute_given_dir_when_file(fakeconfig):
    testdir = os.path.join(fakeconfig.core.homedir, 'test.d')
    with pytest.raises(ValueError):
        fakeconfig.fake.af_fileattr = testdir


def test_fileattribute_given_file_when_dir(fakeconfig):
    testfile = os.path.join(fakeconfig.core.homedir, 'test.tmp')
    with pytest.raises(ValueError):
        fakeconfig.fake.ad_fileattr = testfile


def test_configparser_env_priority_over_file(monkeypatch, fakeconfig):
    monkeypatch.setenv('SOPEL_CORE_OWNER', 'not_dgw')
    assert fakeconfig.core.owner == 'not_dgw'


def test_configparser_multi_lines(multi_fakeconfig):
    # spam
    assert multi_fakeconfig.spam.eggs == [
        'one',
        'two',
        'three',
        'four',
        'and a half',  # no-newline + comma
    ], 'Comma separated line: "four" and "and a half" must be separated'
    assert multi_fakeconfig.spam.bacons == [
        'grilled',
        'burn out',
        'greasy, fat, and tasty',
    ]
    assert multi_fakeconfig.spam.cheeses == [
        'cheddar',
        'reblochon',
        'camembert',
    ]

    assert multi_fakeconfig.spam.channels == TEST_CHANNELS


def test_configparser_multi_env(monkeypatch, multi_fakeconfig):
    monkeypatch.setenv('SOPEL_SPAM_EGGS', 'five, six, seven, eight, and a half')
    monkeypatch.setenv('SOPEL_SPAM_BACONS', 'microwaved\nfreeze in,\n, dry, thin, and disgusting')
    monkeypatch.setenv('SOPEL_SPAM_CHEESES', ' swiss\n  sbrinz\ncottage')
    # Comments not allowed when passing channels via ENV
    monkeypatch.setenv(
        'SOPEL_SPAM_CHANNELS',
        '"#sopel"\n&strange\n*someZnc\n"#public"\n"#frontquote\n&backquote"\n"&bothquoted"\n"*starchan"'
    )

    assert multi_fakeconfig.spam.eggs == [
        'five',
        'six',
        'seven',
        'eight',
        'and a half',  # no-newline + comma
    ]
    assert multi_fakeconfig.spam.bacons == [
        'microwaved',
        'freeze in',
        'dry, thin, and disgusting',
    ]
    assert multi_fakeconfig.spam.cheeses == [
        'swiss',
        'sbrinz',
        'cottage',
    ]
    assert multi_fakeconfig.spam.channels == [
        '#sopel',
        '&strange',
        '*someZnc',
        '#public',
        '"#frontquote',
        '&backquote"',
        '"&bothquoted"',
        '"*starchan"'
    ]


def test_save_unmodified_config(multi_fakeconfig):
    """Assert type attributes are kept as they should be"""
    multi_fakeconfig.save()
    saved_config = config.Config(multi_fakeconfig.filename)
    saved_config.define_section('fake', FakeConfigSection)
    saved_config.define_section('spam', SpamSection)

    # core
    assert saved_config.core.owner == 'dgw'

    # fake
    assert saved_config.fake.valattr is None
    assert saved_config.fake.listattr == []
    assert saved_config.fake.choiceattr is None
    assert saved_config.fake.af_fileattr is None
    assert saved_config.fake.ad_fileattr is None
    assert saved_config.fake.rf_fileattr is None
    assert saved_config.fake.rd_fileattr is None

    # spam
    assert saved_config.spam.eggs == [
        'one',
        'two',
        'three',
        'four',
        'and a half',  # no-newline + comma
    ], 'Comma separated line: "four" and "and a half" must be separated'
    assert saved_config.spam.bacons == [
        'grilled',
        'burn out',
        'greasy, fat, and tasty',
    ]
    assert saved_config.spam.cheeses == [
        'cheddar',
        'reblochon',
        'camembert',
    ]
    assert saved_config.spam.channels == TEST_CHANNELS


def test_save_modified_config(multi_fakeconfig):
    """Assert modified values are restored properly"""
    multi_fakeconfig.fake.choiceattr = 'spam'
    multi_fakeconfig.spam.eggs = [
        'one',
        'two',
    ]
    multi_fakeconfig.spam.cheeses = [
        'camembert, reblochon, and cheddar',
    ]
    multi_fakeconfig.spam.channels = [
        '#sopel',
        '#private',
        '&peculiar',
        '"#startquote',
        '&endquote"',
        '"&quoted"',
    ]

    multi_fakeconfig.save()

    with open(multi_fakeconfig.filename) as fd:
        print(fd.read())  # used for debug purpose if an assert fails

    saved_config = config.Config(multi_fakeconfig.filename)
    saved_config.define_section('fake', FakeConfigSection)
    saved_config.define_section('spam', SpamSection)

    assert saved_config.fake.choiceattr == 'spam'
    assert saved_config.spam.eggs == ['one', 'two']
    assert saved_config.spam.cheeses == [
        'camembert, reblochon, and cheddar',
    ], (
        'ListAttribute with one line only, with commas, must *not* be split '
        'differently from what was expected, i.e. into one (and only one) value'
    )
    assert saved_config.spam.channels == [
        '#sopel',
        '#private',
        '&peculiar',
        '"#startquote',  # start quote without end quote: kept
        '&endquote"',
        '"&quoted"',  # doesn't start with a # so it isn't escaped
    ]


def test_get_defined_sections(multi_fakeconfig):
    assert multi_fakeconfig.parser.has_section('core')
    assert multi_fakeconfig.parser.has_section('fake')
    assert multi_fakeconfig.parser.has_section('spam')
    assert multi_fakeconfig.parser.has_section('somesection')

    results = multi_fakeconfig.get_defined_sections()

    assert len(results) == 3, 'There should be 3 defined sections'

    items = dict(results)
    assert 'core' in items, 'core must be always defined'
    assert 'fake' in items
    assert 'spam' in items
    assert 'somesection' not in items, (
        'somesection was not defined and should not appear as such')



# coding=utf-8
"""Tests for sopel.module decorators

.. important::

    These tests are kept here as a proof that ``sopel.module`` is backward
    compatible up to Sopel 9, when it will be removed.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from sopel import module, tools
from sopel.trigger import PreTrigger, Trigger


TMP_CONFIG = """
[core]
owner = Bar
nick = Sopel
enable = coretasks
"""


FOO_MESSAGE = ':Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
FOO_PRIV_MESSAGE = ':Foo!foo@example.com PRIVMSG Sopel :Hello, world'


@pytest.fixture
def bot(configfactory, botfactory, triggerfactory, ircfactory):
    settings = configfactory('default.cfg', TMP_CONFIG)
    mockbot = botfactory.preloaded(settings)
    mockserver = ircfactory(mockbot)

    bot = triggerfactory.wrapper(mockbot, FOO_MESSAGE)
    mockserver.channel_joined('#Sopel')
    mockserver.join('Foo', '#Sopel')
    mockserver.mode_set('#Sopel', '+v', ['Foo'])

    return bot


@pytest.fixture
def pretrigger():
    return PreTrigger(tools.Identifier('Foo'), FOO_MESSAGE)


@pytest.fixture
def pretrigger_pm():
    return PreTrigger(tools.Identifier('Foo'), FOO_PRIV_MESSAGE)


@pytest.fixture
def trigger_owner(bot):
    line = ':Bar!bar@example.com PRIVMSG #Sopel :Hello, world'
    return Trigger(bot.config, PreTrigger(tools.Identifier('Bar'), line), None)


@pytest.fixture
def trigger_account(bot):
    line = '@account=egg :egg!egg@eg.gs PRIVMSG #Sopel :Hello, world'
    return Trigger(
        bot.config,
        PreTrigger(tools.Identifier('egg'), line),
        None,
        'egg')


@pytest.fixture
def trigger(bot, pretrigger):
    return Trigger(bot.config, pretrigger, None)


@pytest.fixture
def trigger_pm(bot, pretrigger_pm):
    return Trigger(bot.config, pretrigger_pm, None)


def test_unblockable():
    @module.unblockable
    def mock(bot, trigger, match):
        return True
    assert mock.unblockable is True


def test_interval():
    @module.interval(5)
    def mock(bot, trigger, match):
        return True
    assert mock.interval == [5]


def test_interval_args():
    @module.interval(5, 10)
    def mock(bot, trigger, match):
        return True
    assert mock.interval == [5, 10]


def test_interval_multiple():
    @module.interval(5, 10)
    @module.interval(5)
    @module.interval(20)
    def mock(bot, trigger, match):
        return True
    assert mock.interval == [20, 5, 10]


def test_rule():
    @module.rule('.*')
    def mock(bot, trigger, match):
        return True
    assert mock.rule == ['.*']


def test_rule_args():
    @module.rule('.*', r'\d+')
    def mock(bot, trigger, match):
        return True
    assert mock.rule == ['.*', r'\d+']


def test_rule_multiple():
    @module.rule('.*', r'\d+')
    @module.rule('.*')
    @module.rule(r'\w+')
    def mock(bot, trigger, match):
        return True
    assert mock.rule == [r'\w+', '.*', r'\d+']


def test_thread():
    @module.thread(True)
    def mock(bot, trigger, match):
        return True
    assert mock.thread is True


def test_url():
    @module.url('pattern')
    def mock(bot, trigger, match):
        return True
    patterns = [regex.pattern for regex in mock.url_regex]
    assert len(patterns) == 1
    assert 'pattern' in patterns


def test_url_args():
    @module.url('first', 'second')
    def mock(bot, trigger, match):
        return True

    patterns = [regex.pattern for regex in mock.url_regex]
    assert len(patterns) == 2
    assert 'first' in patterns
    assert 'second' in patterns


def test_url_multiple():
    @module.url('first', 'second')
    @module.url('second')
    @module.url('third')
    def mock(bot, trigger, match):
        return True

    patterns = [regex.pattern for regex in mock.url_regex]
    assert len(patterns) == 3
    assert 'first' in patterns
    assert 'second' in patterns
    assert 'third' in patterns


def test_echo():
    # test decorator with parentheses
    @module.echo()
    def mock(bot, trigger, match):
        return True
    assert mock.echo is True

    # test decorator without parentheses
    @module.echo
    def mock(bot, trigger, match):
        return True
    assert mock.echo is True

    # test without decorator
    def mock(bot, trigger, match):
        return True
    # on undecorated callables, the attr only exists after the loader loads them
    # so this cannot `assert mock.echo is False` here
    assert not hasattr(mock, 'echo')


def test_commands():
    @module.commands('sopel')
    def mock(bot, trigger, match):
        return True
    assert mock.commands == ['sopel']
    assert not hasattr(mock, 'rule')


def test_commands_args():
    @module.commands('sopel', 'bot')
    def mock(bot, trigger, match):
        return True
    assert mock.commands == ['sopel', 'bot']
    assert not hasattr(mock, 'rule')


def test_commands_multiple():
    @module.commands('sopel', 'bot')
    @module.commands('bot')
    @module.commands('robot')
    def mock(bot, trigger, match):
        return True
    assert mock.commands == ['robot', 'bot', 'sopel']
    assert not hasattr(mock, 'rule')


def test_nickname_commands():
    @module.nickname_commands('sopel')
    def mock(bot, trigger, match):
        return True
    assert mock.nickname_commands == ['sopel']
    assert not hasattr(mock, 'rule')


def test_nickname_commands_args():
    @module.nickname_commands('sopel', 'bot')
    def mock(bot, trigger, match):
        return True
    assert mock.nickname_commands == ['sopel', 'bot']
    assert not hasattr(mock, 'rule')


def test_nickname_commands_multiple():
    @module.nickname_commands('sopel', 'bot')
    @module.nickname_commands('bot')
    @module.nickname_commands('robot')
    def mock(bot, trigger, match):
        return True
    assert mock.nickname_commands == ['robot', 'bot', 'sopel']
    assert not hasattr(mock, 'rule')


def test_action_commands():
    @module.action_commands('sopel')
    def mock(bot, trigger, match):
        return True
    assert mock.action_commands == ['sopel']
    assert not hasattr(mock, 'intents')
    assert not hasattr(mock, 'rule')


def test_action_commands_args():
    @module.action_commands('sopel', 'bot')
    def mock(bot, trigger, match):
        return True
    assert mock.action_commands == ['sopel', 'bot']
    assert not hasattr(mock, 'intents')
    assert not hasattr(mock, 'rule')


def test_action_commands_multiple():
    @module.action_commands('sopel', 'bot')
    @module.action_commands('bot')
    @module.action_commands('robot')
    def mock(bot, trigger, match):
        return True
    assert mock.action_commands == ['robot', 'bot', 'sopel']
    assert not hasattr(mock, 'intents')
    assert not hasattr(mock, 'rule')


def test_all_commands():
    @module.commands('sopel')
    @module.action_commands('me_sopel')
    @module.nickname_commands('name_sopel')
    def mock(bot, trigger, match):
        return True

    assert mock.commands == ['sopel']
    assert mock.action_commands == ['me_sopel']
    assert mock.nickname_commands == ['name_sopel']
    assert not hasattr(mock, 'intents')
    assert not hasattr(mock, 'rule')


def test_priority():
    @module.priority('high')
    def mock(bot, trigger, match):
        return True
    assert mock.priority == 'high'


def test_event():
    @module.event('301')
    def mock(bot, trigger, match):
        return True
    assert mock.event == ['301']


def test_event_args():
    @module.event('301', '302')
    def mock(bot, trigger, match):
        return True
    assert mock.event == ['301', '302']


def test_event_multiple():
    @module.event('301', '302')
    @module.event('301')
    @module.event('466')
    def mock(bot, trigger, match):
        return True
    assert mock.event == ['466', '301', '302']


def test_intent():
    @module.intent('ACTION')
    def mock(bot, trigger, match):
        return True
    assert mock.intents == ['ACTION']


def test_intent_args():
    @module.intent('ACTION', 'OTHER')
    def mock(bot, trigger, match):
        return True
    assert mock.intents == ['ACTION', 'OTHER']


def test_intent_multiple():
    @module.intent('ACTION', 'OTHER')
    @module.intent('OTHER')
    @module.intent('PING',)
    def mock(bot, trigger, match):
        return True
    assert mock.intents == ['PING', 'OTHER', 'ACTION']


def test_rate():
    @module.rate(5)
    def mock(bot, trigger, match):
        return True
    assert mock.rate == 5


def test_require_privmsg(bot, trigger, trigger_pm):
    @module.require_privmsg('Try again in a PM')
    def mock(bot, trigger, match=None):
        return True
    assert mock(bot, trigger) is not True
    assert mock(bot, trigger_pm) is True

    @module.require_privmsg
    def mock_(bot, trigger, match=None):
        return True
    assert mock_(bot, trigger) is not True
    assert mock_(bot, trigger_pm) is True


def test_require_chanmsg(bot, trigger, trigger_pm):
    @module.require_chanmsg('Try again in a channel')
    def mock(bot, trigger, match=None):
        return True
    assert mock(bot, trigger) is True
    assert mock(bot, trigger_pm) is not True

    @module.require_chanmsg
    def mock_(bot, trigger, match=None):
        return True
    assert mock_(bot, trigger) is True
    assert mock_(bot, trigger_pm) is not True


def test_require_account(bot, trigger, trigger_account):
    @module.require_account('You need to authenticate to services first.')
    def mock(bot, trigger, match=None):
        return True
    assert mock(bot, trigger) is not True
    assert mock(bot, trigger_account) is True

    @module.require_account
    def mock_(bot, trigger, match=None):
        return True
    assert mock_(bot, trigger) is not True
    assert mock_(bot, trigger_account) is True


def test_require_privilege(bot, trigger):
    @module.require_privilege(module.VOICE)
    def mock_v(bot, trigger, match=None):
        return True
    assert mock_v(bot, trigger) is True

    @module.require_privilege(module.OP, 'You must be at least opped!')
    def mock_o(bot, trigger, match=None):
        return True
    assert mock_o(bot, trigger) is not True


def test_require_admin(bot, trigger, trigger_owner):
    @module.require_admin('You must be an admin')
    def mock(bot, trigger, match=None):
        return True
    assert mock(bot, trigger) is not True

    @module.require_admin
    def mock_(bot, trigger, match=None):
        return True
    assert mock_(bot, trigger_owner) is True


def test_require_owner(bot, trigger, trigger_owner):
    @module.require_owner('You must be an owner')
    def mock(bot, trigger, match=None):
        return True
    assert mock(bot, trigger) is not True

    @module.require_owner
    def mock_(bot, trigger, match=None):
        return True
    assert mock_(bot, trigger_owner) is True


def test_example(bot, trigger):
    @module.commands('mock')
    @module.example('.mock', 'True')
    def mock(bot, trigger, match=None):
        return True
    assert mock(bot, trigger) is True


def test_output_prefix():
    @module.commands('mock')
    @module.output_prefix('[MOCK] ')
    def mock(bot, trigger, match):
        return True
    assert mock.output_prefix == '[MOCK] '



# coding=utf-8
"""Regression tests"""
from __future__ import absolute_import, division, print_function, unicode_literals

from sopel import coretasks, tools


TMP_CONFIG = """
[core]
owner = testnick
nick = Sopel
enable = coretasks
"""


def test_bot_legacy_permissions(configfactory, botfactory, triggerfactory):
    """
    Make sure permissions match after being updated from both RPL_NAMREPLY
    and RPL_WHOREPLY, #1482
    """
    mockbot = botfactory(configfactory('default.cfg', TMP_CONFIG))
    nick = tools.Identifier("Admin")

    # RPL_NAMREPLY
    mockwrapper = triggerfactory.wrapper(
        mockbot, ":test.example.com 353 Foo = #test :Foo ~@Admin")
    coretasks.handle_names(mockwrapper, mockwrapper._trigger)

    assert '#test' in mockbot.channels
    assert nick in mockbot.channels["#test"].privileges

    assert '#test' in mockbot.privileges
    assert nick in mockbot.privileges["#test"]

    channel_privileges = mockbot.channels["#test"].privileges[nick]
    privileges = mockbot.privileges["#test"][nick]

    assert channel_privileges == privileges

    # RPL_WHOREPLY
    mockwrapper = triggerfactory.wrapper(
        mockbot,
        ":test.example.com 352 Foo #test "
        "~Admin adminhost test.example.com Admin Hr~ :0 Admin")
    coretasks.recv_who(mockwrapper, mockwrapper._trigger)

    channel_privileges = mockbot.channels["#test"].privileges[nick]
    privileges = mockbot.privileges["#test"][nick]

    assert channel_privileges == privileges
    assert mockbot.users.get(nick) is not None



# coding=utf-8
"""
*Availability: 3+, deprecated in 6.2.0*

This web class will be removed in Sopel 8.0. As of Sopel 7.0, non-deprecated
functions are available in a new package, ``sopel.tools.web``.
"""
# Copyright © 2008, Sean B. Palmer, inamidst.com
# Copyright © 2009, Michael Yanovich <yanovich.1@osu.edu>
# Copyright © 2012, Dimitri Molenaars, Tyrope.nl.
# Copyright © 2012-2013, Elad Alfassa, <elad@fedoraproject.org>
# Copyright © 2019, dgw, technobabbl.es
# Licensed under the Eiffel Forum License 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import sys

import requests

from .tools import deprecated

if sys.version_info.major < 3:
    import httplib
else:
    import http.client as httplib

# Imports to facilitate transition from sopel.web to sopel.tools.web
from .tools.web import (  # noqa
    USER_AGENT,
    DEFAULT_HEADERS as default_headers,
    entity,
    decode,
    quote,
    quote_query,
    urlencode_non_ascii,
    iri_to_uri,
    urlencode,
)

__all__ = [
    'USER_AGENT',
    'default_headers',
    'ca_certs',
    'get',
    'head',
    'post',
    'get_urllib_object',
    'decode',
    'entity',
    'iri_to_uri',
    'quote',
    'quote_query',
    'urlencode',
    'urlencode_non_ascii',
    'MockHttpResponse',
]


# Deprecated sopel.web methods are not moved, so they won't be accessible from the
# new module location (to discourage new code from using them).

ca_certs = None
# This doesn't appear to be used anywhere any more, so it lives in the to-be-removed
# version of sopel.web. At some point it was used to cover an SSL edge case, long ago.


# HTTP GET
@deprecated
def get(uri, timeout=20, headers=None, return_headers=False,
        limit_bytes=None, verify_ssl=True, dont_decode=False):  # pragma: no cover
    """Execute an HTTP GET query on `uri`, and return the result. Deprecated.

    `timeout` is an optional argument, which represents how much time we should
    wait before throwing a timeout exception. It defaults to 20, but can be set
    to higher values if you are communicating with a slow web application.
    `headers` is a dict of HTTP headers to send with the request.  If
    `return_headers` is True, return a tuple of (bytes, headers)

    `limit_bytes` is ignored.

    """
    if not uri.startswith('http'):
        uri = "http://" + uri
    if headers is None:
        headers = default_headers
    else:
        tmp = default_headers.copy()
        tmp.update(headers)
        headers = tmp
    u = requests.get(uri, timeout=timeout, headers=headers, verify=verify_ssl)
    bytes = u.content
    u.close()
    headers = u.headers
    if not dont_decode:
        bytes = u.text
    if not return_headers:
        return bytes
    else:
        headers['_http_status'] = u.status_code
        return (bytes, headers)


# Get HTTP headers
@deprecated
def head(uri, timeout=20, headers=None, verify_ssl=True):  # pragma: no cover
    """Execute an HTTP GET query on `uri`, and return the headers. Deprecated.

    `timeout` is an optional argument, which represents how much time we should
    wait before throwing a timeout exception. It defaults to 20, but can be set
    to higher values if you are communicating with a slow web application.

    """
    if not uri.startswith('http'):
        uri = "http://" + uri
    if headers is None:
        headers = default_headers
    else:
        tmp = default_headers.copy()
        tmp.update(headers)
        headers = tmp
    u = requests.get(uri, timeout=timeout, headers=headers, verify=verify_ssl)
    info = u.headers
    u.close()
    return info


# HTTP POST
@deprecated
def post(uri, query, limit_bytes=None, timeout=20, verify_ssl=True, return_headers=False):  # pragma: no cover
    """Execute an HTTP POST query. Deprecated.

    `uri` is the target URI, and `query` is the POST data.

    If `return_headers` is true, returns a tuple of (bytes, headers).

    `limit_bytes` is ignored.

    """
    if not uri.startswith('http'):
        uri = "http://" + uri
    u = requests.post(uri, timeout=timeout, verify=verify_ssl, data=query)
    bytes = u.raw.read(limit_bytes)
    headers = u.headers
    u.close()
    if not return_headers:
        return bytes
    else:
        headers['_http_status'] = u.status_code
        return (bytes, headers)


# solely for use by get_urllib_object()
class MockHttpResponse(httplib.HTTPResponse):
    "Mock HTTPResponse with data that comes from requests."
    def __init__(self, response):
        self.headers = response.headers
        self.status = response.status_code
        self.reason = response.reason
        self.close = response.close
        self.read = response.raw.read
        self.url = response.url

    def geturl(self):
        return self.url


# For internal use in web.py, (modules can use this if they need a urllib
# object they can execute read() on) Both handles redirects and makes sure
# input URI is UTF-8
@deprecated
def get_urllib_object(uri, timeout, headers=None, verify_ssl=True, data=None):  # pragma: no cover
    """Return an HTTPResponse object for `uri` and `timeout` and `headers`. Deprecated

    """

    if headers is None:
        headers = default_headers
    else:
        tmp = default_headers.copy()
        tmp.update(headers)
        headers = tmp
    if data is not None:
        response = requests.post(uri, timeout=timeout, verify=verify_ssl,
                                 data=data, headers=headers)
    else:
        response = requests.get(uri, timeout=timeout, verify=verify_ssl,
                                headers=headers)
    return MockHttpResponse(response)



# coding=utf-8
"""
currency.py - Sopel Currency Conversion Plugin
Copyright 2013, Elsie Powell, embolalia.com
Copyright 2019, Mikkel Jeppesen
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import re
import time

import requests

from sopel import plugin
from sopel.config import types
from sopel.tools import web


PLUGIN_OUTPUT_PREFIX = '[currency] '
FIAT_PROVIDERS = {
    'exchangerate.host': 'https://api.exchangerate.host/latest?base=EUR',
    'fixer.io': '//data.fixer.io/api/latest?base=EUR&access_key={}',
}
CRYPTO_URL = 'https://api.coingecko.com/api/v3/exchange_rates'
EXCHANGE_REGEX = re.compile(r'''
    ^(\d+(?:\.\d+)?)                                            # Decimal number
    \s*([a-zA-Z]{3})                                            # 3-letter currency code
    \s+(?:in|as|of|to)\s+                                       # preposition
    (([a-zA-Z]{3}$)|([a-zA-Z]{3})\s)+$                          # one or more 3-letter currency code
''', re.VERBOSE)
LOGGER = logging.getLogger(__name__)
UNSUPPORTED_CURRENCY = "Sorry, {} isn't currently supported."
UNRECOGNIZED_INPUT = "Sorry, I didn't understand the input."

rates = {}
rates_updated = 0.0


class CurrencySection(types.StaticSection):
    fiat_provider = types.ChoiceAttribute('fiat_provider',
                                          list(FIAT_PROVIDERS.keys()),
                                          default='exchangerate.host')
    """Which data provider to use (some of which require no API key)"""
    fixer_io_key = types.ValidatedAttribute('fixer_io_key', default=None)
    """API key for Fixer.io (widest currency support)"""
    fixer_use_ssl = types.BooleanAttribute('fixer_use_ssl', default=False)
    """Whether to use SSL (HTTPS) for Fixer API"""
    auto_convert = types.BooleanAttribute('auto_convert', default=False)
    """Whether to convert currencies without an explicit command"""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | auto\\_convert | False | Whether to convert currencies without an explicit command |
    | fiat\\_provider | ratesapi.io | Which data provider to use (some of which require no API key) |
    | fixer\\_io\\_key | 0123456789abcdef0123456789abcdef | API key for Fixer.io (widest currency support) |
    | fixer\\_use\\_ssl | False | Whether to use SSL (HTTPS) for Fixer API (requires paid API access) |
    """
    config.define_section('currency', CurrencySection, validate=False)
    config.currency.configure_setting(
        'fiat_provider',
        'Which exchange rate provider do you want to use?\n'
        'Available choices: {}'.format(
            ', '.join(FIAT_PROVIDERS.keys())
        ))
    if config.currency.fiat_provider == 'fixer.io':
        config.currency.configure_setting('fixer_io_key', 'API key for Fixer.io:')
        config.currency.configure_setting('fixer_use_ssl', 'Use SSL (paid plans only) for Fixer.io?')
    elif config.currency.fixer_io_key is not None:
        # Must be unset or it will override the chosen fiat_provider
        # TODO: this is temporary for Sopel 7.x; see below
        print('Chosen provider is {}; clearing Fixer.io API key.'.format(config.currency.fiat_provider))
        config.currency.fixer_io_key = None
    config.currency.configure_setting('auto_convert', 'Convert currencies without an explicit command?')


def setup(bot):
    bot.config.define_section('currency', CurrencySection)


class FixerError(Exception):
    """A Fixer.io API Error Exception"""
    def __init__(self, status):
        super(FixerError, self).__init__("FixerError: {}".format(status))


class UnsupportedCurrencyError(Exception):
    """A currency is currently not supported by the API"""
    def __init__(self, currency):
        super(UnsupportedCurrencyError, self).__init__(currency)


def build_reply(amount, base, target, out_string):
    rate_raw = get_rate(base, target)
    rate = float(rate_raw)
    result = float(rate * amount)

    digits = 0
    # up to 10 (8+2) digits precision when result is less than 1
    # as smaller results need more precision
    while digits < 8 and 1 / 10**digits > result:
        digits += 1

    digits += 2

    out_string += ' {value:,.{precision}f} {currency},'.format(value=result, precision=digits, currency=target)

    return out_string


def exchange(bot, match):
    """Show the exchange rate between two currencies"""
    if not match:
        bot.reply(UNRECOGNIZED_INPUT)
        return

    try:
        update_rates(bot)  # Try and update rates. Rate-limiting is done in update_rates()
    except requests.exceptions.RequestException as err:
        bot.reply("Something went wrong while I was getting the exchange rate.")
        LOGGER.error("Error in GET request: {}".format(err))
        return
    except (KeyError, ValueError) as err:
        bot.reply("Error: Could not update exchange rates. Try again later.")
        LOGGER.error("{} on update_rates".format(
            'Invalid JSON' if type(err).__name__ == 'ValueError' else 'Missing JSON value',
        ))
        return
    except FixerError as err:
        bot.reply('Sorry, something went wrong with Fixer')
        LOGGER.error(err)
        return

    query = match.string

    targets = query.split()
    amount_in = targets.pop(0)
    base = targets.pop(0)
    targets.pop(0)

    # TODO: Use this instead after dropping Python 2 support
    # amount, base, _, *targets = query.split()

    try:
        amount = float(amount_in)
    except ValueError:
        bot.reply(UNRECOGNIZED_INPUT)
        return
    except OverflowError:
        bot.reply("Sorry, input amount was out of range.")
        return

    if not amount:
        bot.reply("Zero is zero, no matter what country you're in.")
        return

    out_string = '{} {} is'.format(amount_in, base.upper())

    unsupported_currencies = []
    for target in targets:
        try:
            out_string = build_reply(amount, base.upper(), target.upper(), out_string)
        except ValueError:
            LOGGER.error("Raw rate wasn't a float")
            return
        except KeyError as err:
            bot.reply("Error: Invalid rates")
            LOGGER.error("No key: {} in json".format(err))
            return
        except UnsupportedCurrencyError as cur:
            unsupported_currencies.append(cur)

    if unsupported_currencies:
        out_string = out_string + ' (unsupported:'
        for target in unsupported_currencies:
            out_string = out_string + ' {},'.format(target)
        out_string = out_string[0:-1] + ')'
    else:
        out_string = out_string[0:-1]

    bot.say(out_string)


def get_rate(base, target):
    base = base.upper()
    target = target.upper()

    if base not in rates:
        raise UnsupportedCurrencyError(base)

    if target not in rates:
        raise UnsupportedCurrencyError(target)

    return (1 / rates[base]) * rates[target]


def update_rates(bot):
    global rates, rates_updated

    # If we have data that is less than 24h old, return
    if time.time() - rates_updated < 24 * 60 * 60:
        LOGGER.debug('Skipping rate update; cache is less than 24h old')
        return

    # Update crypto rates
    LOGGER.debug('Updating crypto rates from %s', CRYPTO_URL)
    response = requests.get(CRYPTO_URL)
    response.raise_for_status()
    rates_crypto = response.json()

    # Update fiat rates
    if bot.config.currency.fixer_io_key is not None:
        LOGGER.debug('Updating fiat rates from Fixer.io')
        if bot.config.currency.fiat_provider != 'fixer.io':
            # Warning about future behavior change
            LOGGER.warning(
                'fiat_provider is set to %s, but Fixer.io API key is present; '
                'using Fixer anyway. In Sopel 8, fiat_provider will take precedence.',
                bot.config.currency.fiat_provider)

        proto = 'https:' if bot.config.currency.fixer_use_ssl else 'http:'
        response = requests.get(
            proto +
            FIAT_PROVIDERS['fixer.io'].format(
                web.quote(bot.config.currency.fixer_io_key)
            )
        )

        if not response.json()['success']:
            raise FixerError('Fixer.io request failed with error: {}'.format(response.json()['error']))
    else:
        LOGGER.debug('Updating fiat rates from %s', bot.config.currency.fiat_provider)
        response = requests.get(FIAT_PROVIDERS[bot.config.currency.fiat_provider])

    response.raise_for_status()
    rates_fiat = response.json()

    rates = rates_fiat['rates']
    rates['EUR'] = 1.0  # Put this here to make logic easier

    eur_btc_rate = 1 / rates_crypto['rates']['eur']['value']

    for rate in rates_crypto['rates']:
        if rate.upper() not in rates:
            rates[rate.upper()] = rates_crypto['rates'][rate]['value'] * eur_btc_rate

    # if an error aborted the operation prematurely, we want the next call to retry updating rates
    # therefore we'll update the stored timestamp at the last possible moment
    rates_updated = time.time()
    LOGGER.debug('Rate update completed')


@plugin.command('cur', 'currency', 'exchange')
@plugin.example('.cur 100 usd in btc cad eur',
                r'100 USD is [\d\.]+ BTC, [\d\.]+ CAD, [\d\.]+ EUR',
                re=True, online=True, vcr=True)
@plugin.example('.cur 100 usd in btc cad eur can aux',
                r'100 USD is [\d\.]+ BTC, [\d\.]+ CAD, [\d\.]+ EUR, \(unsupported: CAN, AUX\)',
                re=True, online=True, vcr=True)
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def exchange_cmd(bot, trigger):
    """Show the exchange rate between two currencies."""
    if not trigger.group(2):
        bot.reply("No search term. Usage: {}cur 100 usd in btc cad eur"
                  .format(bot.config.core.help_prefix))
        return

    match = EXCHANGE_REGEX.match(trigger.group(2))
    exchange(bot, match)


@plugin.rule(EXCHANGE_REGEX)
@plugin.example('100 usd in btc cad eur')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def exchange_re(bot, trigger):
    if bot.config.currency.auto_convert:
        match = EXCHANGE_REGEX.match(trigger)
        exchange(bot, match)



# coding=utf-8
"""The :mod:`sopel.module` sub-module is replaced by :mod:`sopel.plugin`.

.. deprecated:: 7.1

    Use :mod:`sopel.plugin` instead. This will be removed in Sopel 9.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

# Import everything from sopel.plugin at the time of replacement.
# Everything new from this point on must *not* leak here.
# Therefore, don't add anything to this import list. Ever.
from sopel.plugin import (  # noqa
    ADMIN,
    action_commands,
    commands,
    echo,
    event,
    example,
    HALFOP,
    ctcp as _future_ctcp,
    interval,
    nickname_commands,
    NOLIMIT,
    OP,
    OPER,
    output_prefix,
    OWNER,
    priority,
    rate,
    require_account,
    require_admin,
    require_chanmsg,
    require_owner,
    require_privilege,
    require_privmsg,
    rule,
    thread,
    unblockable,
    url,
    VOICE,
)


def intent(*intent_list):
    """Decorate a callable to trigger on intent messages.

    :param str intent_list: one or more intent(s) on which to trigger (really,
                            the only useful value is ``ACTION``)

    .. versionadded:: 5.2.0

    .. deprecated:: 7.1

    .. important::

        This will be removed in Sopel 9, as the IRCv3 intent specification is
        long dead. You can use :func:`@ctcp <sopel.plugin.ctcp>` instead.
    """
    return _future_ctcp(*intent_list)



# coding=utf-8
"""The formatting module includes functions to apply IRC formatting to text.

*Availability: 4.5+*
"""
# Copyright 2014, Elsie Powell, embolalia.com
# Copyright 2019, dgw, technobabbl.es
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import re
import string
import sys


__all__ = [
    # control chars
    'CONTROL_NORMAL',
    'CONTROL_COLOR',
    'CONTROL_HEX_COLOR',
    'CONTROL_BOLD',
    'CONTROL_ITALIC',
    'CONTROL_UNDERLINE',
    'CONTROL_STRIKETHROUGH',
    'CONTROL_MONOSPACE',
    'CONTROL_REVERSE',
    # utility functions
    'color',
    'hex_color',
    'bold',
    'italic',
    'underline',
    'strikethrough',
    'monospace',
    'reverse',
    'plain',
    # utility class
    'colors',
]

if sys.version_info.major >= 3:
    unicode = str

# Color names are as specified at http://www.mirc.com/colors.html

CONTROL_NORMAL = '\x0f'
"""The control code to reset formatting."""
CONTROL_COLOR = '\x03'
"""The control code to start or end color formatting."""
CONTROL_HEX_COLOR = '\x04'
"""The control code to start or end hexadecimal color formatting."""
CONTROL_BOLD = '\x02'
"""The control code to start or end bold formatting."""
CONTROL_ITALIC = '\x1d'
"""The control code to start or end italic formatting."""
CONTROL_UNDERLINE = '\x1f'
"""The control code to start or end underlining."""
CONTROL_STRIKETHROUGH = '\x1e'
"""The control code to start or end strikethrough formatting."""
CONTROL_MONOSPACE = '\x11'
"""The control code to start or end monospace formatting."""
CONTROL_REVERSE = '\x16'
"""The control code to start or end reverse-color formatting."""

CONTROL_FORMATTING = [
    CONTROL_NORMAL,
    CONTROL_COLOR,
    CONTROL_HEX_COLOR,
    CONTROL_BOLD,
    CONTROL_ITALIC,
    CONTROL_UNDERLINE,
    CONTROL_STRIKETHROUGH,
    CONTROL_MONOSPACE,
    CONTROL_REVERSE,
]
"""A list of all control characters expected to appear as formatting."""

CONTROL_NON_PRINTING = [
    '\x00',
    '\x01',
    '\x02',  # CONTROL_BOLD
    '\x03',  # CONTROL_COLOR
    '\x04',  # CONTROL_HEX_COLOR
    '\x05',
    '\x06',
    '\x07',
    '\x08',
    '\x09',
    '\x0a',
    '\x0b',
    '\x0c',
    '\x0d',
    '\x0e',
    '\x0f',  # CONTROL_NORMAL
    '\x10',
    '\x11',  # CONTROL_MONOSPACE
    '\x12',
    '\x13',
    '\x14',
    '\x15',
    '\x16',  # CONTROL_REVERSE
    '\x17',
    '\x18',
    '\x19',
    '\x1a',
    '\x1b',
    '\x1c',
    '\x1d',  # CONTROL_ITALIC
    '\x1e',  # CONTROL_STRIKETHROUGH
    '\x1f',  # CONTROL_UNDERLINE
    '\x7f',
]

# Regex to detect Control Pattern
COLOR_PATTERN = re.escape(CONTROL_COLOR) + r'((\d{1,2},\d{2})|\d{2})?'
HEX_COLOR_PATTERN = '%s(%s)?' % (
    re.escape(CONTROL_HEX_COLOR),
    '|'.join([
        '(' + ','.join([r'[a-fA-F0-9]{6}', r'[a-fA-F0-9]{6}']) + ')',
        r'[a-fA-F0-9]{6}'
    ])
)

PLAIN_PATTERN = '|'.join([
    '(' + COLOR_PATTERN + ')',
    '(' + HEX_COLOR_PATTERN + ')',
])
PLAIN_REGEX = re.compile(PLAIN_PATTERN)


# TODO when we can move to 3.3+ completely, make this an Enum.
class colors:
    WHITE = '00'
    BLACK = '01'
    BLUE = '02'
    NAVY = BLUE
    GREEN = '03'
    RED = '04'
    BROWN = '05'
    MAROON = BROWN
    PURPLE = '06'
    ORANGE = '07'
    OLIVE = ORANGE
    YELLOW = '08'
    LIGHT_GREEN = '09'
    LIME = LIGHT_GREEN
    TEAL = '10'
    LIGHT_CYAN = '11'
    CYAN = LIGHT_CYAN
    LIGHT_BLUE = '12'
    ROYAL = LIGHT_BLUE
    PINK = '13'
    LIGHT_PURPLE = PINK
    FUCHSIA = PINK
    GREY = '14'
    LIGHT_GREY = '15'
    SILVER = LIGHT_GREY

    # Create aliases.
    GRAY = GREY
    LIGHT_GRAY = LIGHT_GREY


def _get_color(color):
    if color is None:
        return None

    # You can pass an int or string of the code
    try:
        color = int(color)
    except ValueError:
        pass
    if isinstance(color, int):
        if color > 99:
            raise ValueError('Can not specify a color above 99.')
        return unicode(color).rjust(2, '0')

    # You can also pass the name of the color
    color_name = color.upper()
    color_dict = colors.__dict__
    try:
        return color_dict[color_name]
    except KeyError:
        raise ValueError('Unknown color name {}'.format(color))


def color(text, fg=None, bg=None):
    """Return the text, with the given colors applied in IRC formatting.

    :param str text: the text to format
    :param mixed fg: the foreground color
    :param mixed bg: the background color

    The color can be a string of the color name, or an integer in the range
    0-99. The known color names can be found in the :class:`colors` class of
    this module.
    """
    if not fg and not bg:
        return text

    fg = _get_color(fg)
    bg = _get_color(bg)

    if not bg:
        text = ''.join([CONTROL_COLOR, fg, text, CONTROL_COLOR])
    else:
        text = ''.join([CONTROL_COLOR, fg, ',', bg, text, CONTROL_COLOR])
    return text


def _get_hex_color(color):
    if color is None:
        return None

    try:
        color = color.upper()
        if not all(c in string.hexdigits for c in color):
            raise AttributeError
    except AttributeError:
        raise ValueError('Hexadecimal color value must be passed as string.')

    if len(color) == 3:
        return ''.join([c * 2 for c in color])
    elif len(color) == 6:
        return color
    else:  # invalid length
        raise ValueError('Hexadecimal color value must have either 3 or 6 digits.')


def hex_color(text, fg=None, bg=None):
    """Return the text, with the given colors applied in IRC formatting.

    :param str text: the text to format
    :param str fg: the foreground color
    :param str bg: the background color

    The color can be provided with a string of either 3 or 6 hexadecimal digits.
    As in CSS, 3-digit colors will be interpreted as if they were 6-digit colors
    with each digit repeated (e.g. color ``c90`` is identical to ``cc9900``). Do
    not include the leading ``#`` symbol.

    .. note::
        This is a relatively new IRC formatting convention. Use only when you
        can afford to have its meaning lost, as not many clients support it yet.
    """
    if not fg and not bg:
        return text

    fg = _get_hex_color(fg)
    bg = _get_hex_color(bg)

    if not bg:
        text = ''.join([CONTROL_HEX_COLOR, fg, text, CONTROL_HEX_COLOR])
    else:
        text = ''.join([CONTROL_HEX_COLOR, fg, ',', bg, text, CONTROL_HEX_COLOR])
    return text


def bold(text):
    """Return the text, with bold IRC formatting.

    :param str text: the text to format
    """
    return ''.join([CONTROL_BOLD, text, CONTROL_BOLD])


def italic(text):
    """Return the text, with italic IRC formatting.

    :param str text: the text to format
    """
    return ''.join([CONTROL_ITALIC, text, CONTROL_ITALIC])


def underline(text):
    """Return the text, with underline IRC formatting.

    :param str text: the text to format
    """
    return ''.join([CONTROL_UNDERLINE, text, CONTROL_UNDERLINE])


def strikethrough(text):
    """Return the text, with strikethrough IRC formatting.

    :param str text: the text to format

    .. note::
        This is a relatively new IRC formatting convention. Use only when you
        can afford to have its meaning lost, as not many clients support it yet.
    """
    return ''.join([CONTROL_STRIKETHROUGH, text, CONTROL_STRIKETHROUGH])


def monospace(text):
    """Return the text, with monospace IRC formatting.

    :param str text: the text to format

    .. note::
        This is a relatively new IRC formatting convention. Use only when you
        can afford to have its meaning lost, as not many clients support it yet.
    """
    return ''.join([CONTROL_MONOSPACE, text, CONTROL_MONOSPACE])


def reverse(text):
    """Return the text, with reverse-color IRC formatting.

    :param str text: the text to format

    .. note::
        This code isn't super well supported, and its behavior even in clients
        that understand it (e.g. mIRC) can be unpredictable. Use it carefully.
    """
    return ''.join([CONTROL_REVERSE, text, CONTROL_REVERSE])


def plain(text):
    """Return the text without any IRC formatting.

    :param str text: text with potential IRC formatting control code(s)
    :rtype: str
    """
    if '\x03' in text or '\x04' in text:
        text = PLAIN_REGEX.sub('', text)
    return ''.join(c for c in text if ord(c) >= 0x20 and c != '\x7F')



# coding=utf-8
"""
bugzilla.py - Sopel Bugzilla Plugin
Copyright 2013-2015, Embolalia, embolalia.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import re

import requests
import xmltodict

from sopel import plugin, plugins
from sopel.config import types

LOGGER = logging.getLogger(__name__)


class BugzillaSection(types.StaticSection):
    domains = types.ListAttribute('domains')
    """A list of Bugzilla issue tracker domains from which to get information."""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | domains | bugzilla.redhat.com,bugzilla.mozilla.org | A list of Bugzilla issue tracker domains |
    """
    config.define_section('bugzilla', BugzillaSection)
    config.bugzilla.configure_setting(
        'domains',
        'Enter the domains of the Bugzillas you want extra information '
        'from (e.g. bugzilla.gnome.org)'
    )


def setup(bot):
    bot.config.define_section('bugzilla', BugzillaSection)


def _bugzilla_loader(settings):
    if not settings.bugzilla.domains:
        raise plugins.exceptions.PluginSettingsError(
            'Bugzilla URL callback requires '
            '"bugzilla.domains" to be configured; check your config file.')

    domain_pattern = '|'.join(
        re.escape(domain)
        for domain in settings.bugzilla.domains)

    pattern = (
        r'https?://(%s)'
        r'(/show_bug.cgi\?\S*?)'
        r'(id=\d+).*'
    ) % domain_pattern

    return [re.compile(pattern)]


@plugin.url_lazy(_bugzilla_loader)
@plugin.output_prefix('[BUGZILLA] ')
def show_bug(bot, trigger, match=None):
    """Show information about a Bugzilla bug."""
    url = 'https://%s%sctype=xml&%s' % trigger.groups()
    data = requests.get(url).content
    bug = xmltodict.parse(data).get('bugzilla').get('bug')
    error = bug.get('@error', None)  # error="NotPermitted"

    if error:
        LOGGER.warning('Bugzilla error: %s' % error)
        bot.say('Unable to get information for '
                'linked bug (%s)' % error)
        return

    message = ('%s | Product: %s | Component: %s | Version: %s | ' +
               'Importance: %s |  Status: %s | Assigned to: %s | ' +
               'Reported: %s | Modified: %s')

    resolution = bug.get('resolution')
    if resolution is not None:
        status = bug.get('bug_status') + ' ' + resolution
    else:
        status = bug.get('bug_status')

    assigned_to = bug.get('assigned_to')
    if isinstance(assigned_to, dict):
        assigned_to = assigned_to.get('@name')

    message = message % (
        bug.get('short_desc'), bug.get('product'),
        bug.get('component'), bug.get('version'),
        (bug.get('priority') + ' ' + bug.get('bug_severity')),
        status, assigned_to, bug.get('creation_ts'),
        bug.get('delta_ts'))
    bot.say(message)



# coding=utf-8
"""Triggers are how Sopel tells callables about their runtime context."""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import re
import sys

from sopel import formatting, tools
from sopel.tools import web


__all__ = [
    'PreTrigger',
    'Trigger',
]

if sys.version_info.major >= 3:
    unicode = str
    basestring = str


class PreTrigger(object):
    """A parsed raw message from the server.

    :param str own_nick: the bot's own IRC nickname
    :param str line: the full line from the server
    :param tuple url_schemes: allowed schemes for URL detection

    At the :class:`PreTrigger` stage, the line has not been matched against any
    rules yet. This is what Sopel uses to perform matching.

    ``own_nick`` is needed to correctly parse who sent the line (Sopel or
    someone else).

    ``line`` can also be a simulated echo-message, useful if the connected
    server does not support the capability.

    .. py:attribute:: args

        The IRC command's arguments.

        These are split on spaces, per the IRC protocol.

    .. py:attribute:: event

        The IRC command name or numeric value.

        See :class:`sopel.tools.events` for a built-in mapping of names to
        numeric values.

    .. py:attribute:: host

        The sender's hostname.

    .. py:attribute:: hostmask

        The sender's hostmask, as sent by the IRC server.

    .. py:attribute:: line

        The raw line received from the server.

    .. py:attribute:: nick

        The nickname that sent this command.

    .. py:attribute:: sender

        Channel name or query where this message was received.

    .. py:attribute:: tags

        Any IRCv3 message tags attached to the line, as a :class:`dict`.

    .. py:attribute:: text

        The last argument of the IRC command.

        If the line contains ``:``, :attr:`text` will be the text following it.

        For lines that do *not* contain ``:``, :attr:`text` will be the last
        argument in :attr:`args` instead.

    .. py:attribute:: urls
        :type: tuple

        List of URLs found in the :attr:`text`.

        This is for ``PRIVMSG`` and ``NOTICE`` messages only. For other
        messages, this will be an empty ``tuple``.

    .. py:attribute:: plain

        The last argument of the IRC command with control codes stripped.

    .. py:attribute:: time

        The time when the message was received.

        If the IRC server sent a message tag indicating when *it* received the
        message, that is used instead of the time when Sopel received it.

    .. py:attribute:: user

        The sender's local username.

    """
    component_regex = re.compile(r'([^!]*)!?([^@]*)@?(.*)')
    intent_regex = re.compile('\x01(\\S+) ?(.*)\x01')

    def __init__(self, own_nick, line, url_schemes=None):
        line = line.strip('\r\n')
        self.line = line
        self.urls = tuple()
        self.plain = ''

        # Break off IRCv3 message tags, if present
        self.tags = {}
        if line.startswith('@'):
            tagstring, line = line.split(' ', 1)
            for tag in tagstring[1:].split(';'):
                tag = tag.split('=', 1)
                if len(tag) > 1:
                    self.tags[tag[0]] = tag[1]
                else:
                    self.tags[tag[0]] = None

        self.time = datetime.datetime.utcnow()
        if 'time' in self.tags:
            try:
                self.time = datetime.datetime.strptime(self.tags['time'], '%Y-%m-%dT%H:%M:%S.%fZ')
            except ValueError:
                pass  # Server isn't conforming to spec, ignore the server-time

        # Grabs hostmask from line.
        # Example: line = ':Sopel!foo@bar PRIVMSG #sopel :foobar!'
        #          print(hostmask)  # Sopel!foo@bar
        # All lines start with ":" except PING.
        if line.startswith(':'):
            self.hostmask, line = line[1:].split(' ', 1)
        else:
            self.hostmask = None

        # Parses the line into a list of arguments.
        # Some events like MODE don't have a secondary string argument, i.e. no ' :' inside the line.
        # Example 1:  line = ':nick!ident@domain PRIVMSG #sopel :foo bar!'
        #             print(text)    # 'foo bar!'
        #             print(argstr)  # ':nick!ident@domain PRIVMSG #sopel'
        #             print(args)    # [':nick!ident@domain', 'PRIVMSG', '#sopel', 'foo bar!']
        # Example 2:  line = 'irc.libera.chat MODE Sopel +i'
        #             print(text)    # '+i'
        #             print(args)    # ['irc.libera.chat', 'MODE', 'Sopel', '+i']
        if ' :' in line:
            argstr, self.text = line.split(' :', 1)
            self.args = argstr.split(' ')
            self.args.append(self.text)
        else:
            self.args = line.split(' ')
            self.text = self.args[-1]

        self.event = self.args[0]
        self.args = self.args[1:]
        components = PreTrigger.component_regex.match(self.hostmask or '').groups()
        self.nick, self.user, self.host = components
        self.nick = tools.Identifier(self.nick)

        # If we have arguments, the first one is the sender
        # Unless it's a QUIT event
        if self.args and self.event != 'QUIT':
            target = tools.Identifier(self.args[0])
        else:
            target = None

        # Unless we're messaging the bot directly, in which case that second
        # arg will be our bot's name.
        if target and target.lower() == own_nick.lower():
            target = self.nick
        self.sender = target

        # Parse CTCP into a form consistent with IRCv3 intents
        if self.event == 'PRIVMSG' or self.event == 'NOTICE':
            intent_match = PreTrigger.intent_regex.match(self.args[-1])
            if intent_match:
                intent, message = intent_match.groups()
                self.tags['intent'] = intent
                self.args[-1] = message or ''

            # Search URLs after CTCP parsing
            self.urls = tuple(
                web.search_urls(self.args[-1], schemes=url_schemes))

        # Populate account from extended-join messages
        if self.event == 'JOIN' and len(self.args) == 3:
            # Account is the second arg `...JOIN #Sopel account :realname`
            self.tags['account'] = self.args[1]

        # get plain text message
        if self.args:
            self.plain = formatting.plain(self.args[-1])


class Trigger(unicode):
    """A line from the server, which has matched a callable's rules.

    :param config: Sopel's current configuration settings object
    :type config: :class:`~sopel.config.Config`
    :param message: the message that matched the callable
    :type message: :class:`PreTrigger`
    :param match: what *in* the message matched
    :type match: :ref:`Match object <match-objects>`
    :param str account: services account name of the ``message``'s sender
                        (optional; only applies on networks with the
                        ``account-tag`` capability enabled)

    A :class:`Trigger` object itself can be used as a string; when used in
    this way, it represents the matching line's full text.

    The ``match`` is based on the matching regular expression rule; Sopel's
    command decorators auto-generate rules containing specific numbered groups
    that are documented separately. (See :attr:`group` below.)

    Note that CTCP messages (``PRIVMSG``\\es and ``NOTICE``\\es which start
    and end with ``'\\x01'``) will have the ``'\\x01'`` bytes stripped, and
    the command (e.g. ``ACTION``) placed mapped to the ``'intent'`` key in
    :attr:`Trigger.tags`.
    """
    sender = property(lambda self: self._pretrigger.sender)
    """Where the message arrived from.

    :type: :class:`~.tools.Identifier`

    This will be a channel name for "regular" (channel) messages, or the nick
    that sent a private message.

    You can check if the trigger comes from a channel or a nick with its
    :meth:`~sopel.tools.Identifier.is_nick` method::

        if trigger.sender.is_nick():
            # message sent from a private message
        else:
            # message sent from a channel
    """
    time = property(lambda self: self._pretrigger.time)
    """When the message was received.

    :type: naïve :class:`~datetime.datetime` object (no timezone)

    If the IRC server supports ``server-time``, :attr:`time` will give that
    value. Otherwise, :attr:`time` will use the time when the message was
    received by Sopel. In both cases, this time is in UTC.
    """
    raw = property(lambda self: self._pretrigger.line)
    """The entire raw IRC message, as sent from the server.

    :type: str

    This includes the CTCP ``\\x01`` bytes and command, if they were included.
    """
    is_privmsg = property(lambda self: self._is_privmsg)
    """Whether the message was triggered in a private message.

    :type: bool

    ``True`` if the trigger was received in a private message; ``False`` if it
    came from a channel.
    """
    hostmask = property(lambda self: self._pretrigger.hostmask)
    """Hostmask of the person who sent the message as ``<nick>!<user>@<host>``.

    :type: str
    """
    user = property(lambda self: self._pretrigger.user)
    """Local username of the person who sent the message.

    :type: str
    """
    nick = property(lambda self: self._pretrigger.nick)
    """The nickname who sent the message.

    :type: :class:`~.tools.Identifier`
    """
    host = property(lambda self: self._pretrigger.host)
    """The hostname of the person who sent the message.

    :type: str
    """
    event = property(lambda self: self._pretrigger.event)
    """The IRC event which triggered the message.

    :type: str

    Most plugin :func:`callables <callable>` primarily need to deal with
    ``PRIVMSG``. Other event types like ``NOTICE``, ``NICK``, ``TOPIC``,
    ``KICK``, etc. must be requested using :func:`.plugin.event`.
    """
    ctcp = property(lambda self: self.tags.get('intent', None))
    """The CTCP command (if any).

    :type: str

    Common CTCP commands are ``ACTION``, ``VERSION``, and ``TIME``. Other
    commands include ``USERINFO``, ``PING``, and various ``DCC`` operations.

    .. versionadded:: 7.1

    .. important::

        Use this attribute instead of the ``intent`` tag in :attr:`tags`.
        Message intents never made it past the IRCv3 draft stage, and Sopel will
        drop support for them in a future release.
    """
    match = property(lambda self: self._match)
    """The :ref:`Match object <match-objects>` for the triggering line.

    :type: :ref:`Match object <match-objects>`
    """
    group = property(lambda self: self._match.group)
    """The ``group()`` function of the :attr:`match` attribute.

    :type: :term:`method`
    :rtype: str

    Any regular-expression :func:`rules <.plugin.rule>` attached to the
    triggered :func:`callable` may define numbered or named groups that can be
    retrieved through this property.

    Sopel's command decorators each define a predetermined set of numbered
    groups containing fragments of the line that plugins commonly use.

    .. seealso::

       For more information on predefined numbered match groups in commands,
       see :func:`.plugin.command`, :func:`.plugin.action_command`, and
       :func:`.plugin.nickname_command`.

       Also see Python's :meth:`re.Match.group` documentation for details
       about this method's behavior.

    """
    groups = property(lambda self: self._match.groups)
    """The ``groups()`` function of the :attr:`match` attribute.

    :type: :term:`method`
    :rtype: tuple

    See Python's :meth:`re.Match.groups` documentation for details.
    """
    groupdict = property(lambda self: self._match.groupdict)
    """The ``groupdict()`` function of the :attr:`match` attribute.

    :type: :term:`method`
    :rtype: dict

    See Python's :meth:`re.Match.groupdict` documentation for details.
    """
    args = property(lambda self: self._pretrigger.args)
    """A list containing each of the arguments to an event.

    :type: list[str]

    These are the strings passed between the event name and the colon. For
    example, when setting ``mode -m`` on the channel ``#example``, args would
    be ``['#example', '-m']``
    """
    urls = property(lambda self: self._pretrigger.urls)
    """A tuple containing all URLs found in the text.

    :type: tuple

    URLs are listed only for ``PRIVMSG`` or a ``NOTICE``, otherwise this is
    an empty tuple.
    """
    plain = property(lambda self: self._pretrigger.plain)
    """The text without formatting control codes.

    :type: str

    This is the text of the trigger object without formatting control codes.
    """
    tags = property(lambda self: self._pretrigger.tags)
    """A map of the IRCv3 message tags on the message.

    :type: dict
    """
    admin = property(lambda self: self._admin)
    """Whether the triggering :attr:`nick` is one of the bot's admins.

    :type: bool

    ``True`` if the triggering :attr:`nick` is a Sopel admin; ``False`` if not.

    Note that Sopel's :attr:`~.config.core_section.CoreSection.owner` is also
    considered to be an admin.
    """
    owner = property(lambda self: self._owner)
    """Whether the :attr:`nick` which triggered the command is the bot's owner.

    :type: bool

    ``True`` if the triggering :attr:`nick` is Sopel's owner; ``False`` if not.
    """
    account = property(lambda self: self.tags.get('account') or self._account)
    """The services account name of the user sending the message.

    :type: str or None

    Note: This is only available if the ``account-tag`` capability or *both*
    the ``account-notify`` and ``extended-join`` capabilities are available on
    the connected IRC network. If this is not the case, or if the user sending
    the message isn't logged in to services, this property will be ``None``.
    """

    def __new__(cls, config, message, match, account=None):
        self = unicode.__new__(cls, message.args[-1] if message.args else '')
        self._account = account
        self._pretrigger = message
        self._match = match
        self._is_privmsg = message.sender and message.sender.is_nick()

        def match_host_or_nick(pattern):
            pattern = tools.get_hostmask_regex(pattern)
            return bool(
                pattern.match(self.nick) or
                pattern.match('@'.join((self.nick, self.host)))
            )

        if config.core.owner_account:
            self._owner = config.core.owner_account == self.account
        else:
            self._owner = match_host_or_nick(config.core.owner)
        self._admin = (
            self._owner or
            self.account in config.core.admin_accounts or
            any(match_host_or_nick(item) for item in config.core.admins)
        )

        return self
